#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "../utilities/openglheader.h"

#include "../utilities/utilities.h"
#include "../utilities/mygltext.h"
#include "lights.h"
#include "trans.h"
#include "app1d.h"
#include "app1dproc.h"

#define ANGULAR_VELOCITY (0.25*PI)

GLuint  program_id[5];
GLuint  icos_vao, icos_vbo[4];
int     option = 2;
char    enlight = true;
char    tesselate = false;
TransBl trans;
LightBl light;
int     win_width, win_height;
float   left, right, bottom, top, near, far;
float   model_rot_axis[3] = {0.0,1.0,0.0};
double  model_rot_angle = 0.0;
float   viewer_rvec[3] = {1.0,0.0,0.0};
double  viewer_rangle = 0.0;
const float viewer_pos0[4] = {0.0,0.0,10.0,1.0};
myTextObject *vptext;
myFont  *fonts[2], *font;

void LoadMyShaders ( void )
{
  static const char *filename[13] =
    { "app1d0.vert.glsl", "app1d1.vert.glsl", "app1d2.vert.glsl",
      "app1d2.tesc.glsl", "app1d3.tesc.glsl",
      "app1d2.tese.glsl", "app1d3.tese.glsl", "app1d4.tese.glsl",
      "app1d1.geom.glsl", "app1d2.geom.glsl",
      "app1d0.frag.glsl", "app1d1.frag.glsl", "app1d3.frag.glsl" };
  static const GLuint shtype[13] =
    { GL_VERTEX_SHADER, GL_VERTEX_SHADER, GL_VERTEX_SHADER,
      GL_TESS_CONTROL_SHADER, GL_TESS_CONTROL_SHADER,
      GL_TESS_EVALUATION_SHADER, GL_TESS_EVALUATION_SHADER,
      GL_TESS_EVALUATION_SHADER, GL_GEOMETRY_SHADER, GL_GEOMETRY_SHADER, 
      GL_FRAGMENT_SHADER, GL_FRAGMENT_SHADER, GL_FRAGMENT_SHADER };
  GLuint shader_id[13], sh[5];
  int    i;

  for ( i = 0; i < 13; i++ )
    shader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  sh[0] = shader_id[0];  sh[1] = shader_id[10];
  program_id[0] = LinkShaderProgram ( 2, sh, "0" );
  sh[0] = shader_id[1];  sh[1] = shader_id[8];  sh[2] = shader_id[11];
  program_id[1] = LinkShaderProgram ( 3, sh, "1" );
  sh[0] = shader_id[2];  sh[1] = shader_id[3];  sh[2] = shader_id[5];
  sh[3] = shader_id[10];
  program_id[2] = LinkShaderProgram ( 4, sh, "2" );
  sh[0] = shader_id[2];  sh[1] = shader_id[4];  sh[2] = shader_id[6];
  sh[3] = shader_id[10];
  program_id[3] = LinkShaderProgram ( 4, sh, "3" );
  sh[0] = shader_id[2];  sh[1] = shader_id[4];  sh[2] = shader_id[7];
  sh[3] = shader_id[9];  sh[4] = shader_id[12];
  program_id[4] = LinkShaderProgram ( 5, sh, "4" );
  GetAccessToTransBlockUniform ( program_id[1] );
  AttachUniformTransBlockToBP ( program_id[0] );
  AttachUniformTransBlockToBP ( program_id[2] );
  AttachUniformTransBlockToBP ( program_id[3] );
  AttachUniformTransBlockToBP ( program_id[4] );
  GetAccessToLightBlockUniform ( program_id[1] );
  AttachUniformLightBlockToBP ( program_id[4] );
  for ( i = 0; i < 13; i++ )
    glDeleteShader ( shader_id[i] );
  ExitIfGLError ( "LoadMyShaders" );
} /*LoadMyShaders*/

void SetupModelMatrix ( float axis[3], float angle )
{
  GLfloat mm[16];

  M4x4RotateVf ( mm, axis[0], axis[1], axis[2], angle );
  LoadMMatrix ( &trans, mm );
} /*SetupModelMatrix*/

void NotifyViewerPos ( void )
{
  GLchar s[60];

  sprintf ( s, "x = %5.2f, y = %5.2f, z = %5.2f",
            trans.eyepos[0], trans.eyepos[1], trans.eyepos[2] );
  SetTextObjectContents ( vptext, s, 0, win_height-1, font );
} /*NotifyViewerPos*/

void InitViewMatrix ( void )
{
  memcpy ( trans.eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  M4x4Translatef ( trans.vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  LoadVPMatrix ( &trans );
  NotifyViewerPos ();
} /*InitViewMatrix*/

void RotateViewer ( double delta_xi, double delta_eta )
{
  float   vi[3], lgt, vk[3];
  double  angi, angk;

  if ( delta_xi == 0 && delta_eta == 0 )
    return;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  vi[0] = (float)delta_eta*(right-left)/(float)win_height;
  vi[1] = (float)delta_xi*(top-bottom)/(float)win_width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  V3CompRotationsf ( vk, &angk, viewer_rvec, viewer_rangle, vi, angi );
  memcpy ( viewer_rvec, vk, 3*sizeof(float) );
  viewer_rangle = angk;
  M4x4RotateVfv ( trans.vm, viewer_rvec, -viewer_rangle );
  M4x4MultMTVf ( trans.eyepos, trans.vm, viewer_pos0 );
  M4x4InvTranslateMfv ( trans.vm, viewer_pos0 );
  LoadVPMatrix ( &trans );
  NotifyViewerPos ();
} /*RotateViewer*/

void ConstructIcosahedronVAO ( void )
{
#define A 0.52573115
#define B 0.85065085
  static const GLfloat vertpos[12][3] =
    {{ -A,0.0, -B},{  A,0.0, -B},{0.0, -B, -A},{ -B, -A,0.0},
     { -B,  A,0.0},{0.0,  B, -A},{  A,0.0,  B},{ -A,0.0,  B},
     {0.0, -B,  A},{  B, -A,0.0},{  B,  A,0.0},{0.0,  B,  A}};
  static const GLubyte vertcol[12][3] =
    {{255,0,0},{255,127,0},{255,255,0},{127,255,0},{0,255,0},{0,255,127},
     {0,255,255},{0,127,255},{0,0,255},{127,0,255},{255,0,255},{255,0,127}};
  static const GLubyte vertind[96] =
     { 0, 1, 2, 0, 3, 4, 0, 5, 1, 9, 2, 8, 3, /* lamana, od 0 */
       7, 4, 11, 5, 10, 9, 6, 8, 7, 6, 11, 7,
       1, 10, 6,                              /* lamana, od 25 */
       2, 3, 4, 5, 8, 9, 10, 11,              /* 4 odcinki, od 28 */
       0, 1, 2, 0, 2, 3, 0, 3, 4, 0, 4, 5,    /* trojkatne platy, od 36 */
       0, 5, 1, 6, 7, 8, 6, 8, 9, 6, 9, 10,
       6, 10, 11, 6, 11, 7, 1, 9, 2, 9, 8, 2,
       2, 8, 3, 8, 7, 3, 3, 7, 4, 7, 11, 4,
       4, 11, 5, 11, 10, 5, 10, 1, 5, 10, 9, 1};

  glGenVertexArrays ( 1, &icos_vao );
  glBindVertexArray ( icos_vao );
  glGenBuffers ( 3, icos_vbo );
  glBindBuffer ( GL_ARRAY_BUFFER, icos_vbo[0] );
  glBufferData ( GL_ARRAY_BUFFER,
                 12*3*sizeof(GLfloat), vertpos, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 0 );
  glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE,
                          3*sizeof(GLfloat), (GLvoid*)0 );
  glBindBuffer ( GL_ARRAY_BUFFER, icos_vbo[1] );
  glBufferData ( GL_ARRAY_BUFFER,
                 12*3*sizeof(GLubyte), vertcol, GL_STATIC_DRAW );
  glEnableVertexAttribArray ( 1 );
  glVertexAttribPointer ( 1, 3, GL_UNSIGNED_BYTE, GL_TRUE,
                          3*sizeof(GLubyte), (GLvoid*)0 );
  glBindBuffer ( GL_ELEMENT_ARRAY_BUFFER, icos_vbo[2] );
  glBufferData ( GL_ELEMENT_ARRAY_BUFFER,
                 96*sizeof(GLubyte), vertind, GL_STATIC_DRAW );
  glBindVertexArray ( 0 );
  ExitIfGLError ( "ConstructIcosahedronVAO" );
} /*ConstructIcosahedronVAO*/

void DrawIcosahedron ( int opt, char enlight )
{
  glBindVertexArray ( icos_vao );
  switch ( opt ) {
case 0:    /* wierzcholki */
    glUseProgram ( program_id[0] );
    glPointSize ( 5.0 );
    glDrawArrays ( GL_POINTS, 0, 12 );
    break;
case 1:    /* krawedzie */
    glUseProgram ( program_id[0] );
    glDrawElements ( GL_LINE_STRIP, 25,
                     GL_UNSIGNED_BYTE, (GLvoid*)0 );
    glDrawElements ( GL_LINE_STRIP, 3,
                     GL_UNSIGNED_BYTE, (GLvoid*)(25*sizeof(GLubyte)) );
    glDrawElements ( GL_LINES, 8, 
                     GL_UNSIGNED_BYTE, (GLvoid*)(28*sizeof(GLubyte)) );
    break;
default:   /* sciany */
    glUseProgram ( program_id[enlight ? 1 : 0] );
    glDrawElements ( GL_TRIANGLES, 60,
                     GL_UNSIGNED_BYTE, (GLvoid*)(36*sizeof(GLubyte)) );
    break;
  }
  glBindVertexArray ( 0 );
  ExitIfGLError ( "DrawIcosahedron" );
} /*DrawIcosahedron*/

void DrawTessIcos ( int opt, char enlight )
{
  glBindVertexArray ( icos_vao );
  switch ( opt ) {
case 0:
    break;
case 1:
    glUseProgram ( program_id[2] );
    glPatchParameteri ( GL_PATCH_VERTICES, 2 );
    glDrawElements ( GL_PATCHES, 24, GL_UNSIGNED_BYTE, (GLvoid*)0 );
    glDrawElements ( GL_PATCHES, 26, GL_UNSIGNED_BYTE, (GLvoid*)(1*sizeof(GLubyte)) );
    glDrawElements ( GL_PATCHES, 10, GL_UNSIGNED_BYTE, (GLvoid*)(26*sizeof(GLubyte)) );
    break;
default:
    glUseProgram ( program_id[enlight ? 4 : 3] );
    glPatchParameteri ( GL_PATCH_VERTICES, 3 );
    glDrawElements ( GL_PATCHES, 60, GL_UNSIGNED_BYTE, (GLvoid*)(36*sizeof(GLubyte)) );
  }
  glBindVertexArray ( 0 );
  ExitIfGLError ( "DrawTessIcos" );
} /*DrawTessIcos*/

void InitLights ( void )
{
  GLfloat amb0[4] = { 0.2, 0.2, 0.3 };
  GLfloat dir0[4] = { 0.8, 0.8, 0.8 };
  GLfloat pos0[4] = { 0.0, 1.0, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  SetLightAmbient ( &light, 0, amb0 );
  SetLightDirect ( &light, 0, dir0 );
  SetLightPosition ( &light, 0, pos0 );
  SetLightAttenuation ( &light, 0, atn0 );
  SetLightOnOff ( &light, 0, 1 );
} /*InitLights*/

void InitMyWorld ( int argc, char *argv[], int width, int height )
{
  static GLfloat fg[4] = { 0.0, 0.0, 1.0, 1.0 };
  static GLfloat bk[4] = { 0.0, 0.0, 0.0, 0.0 };

  LoadMyShaders ();
  LoadTextShaders ();
  TimerInit ();
  font = fonts[0] = NewFont18x10 ();
  fonts[1] = NewFont12x6 ();
  vptext = NewTextObject ( 60 );
  SetTextForeground ( fg );
  SetTextBackground ( bk );
  memset ( &trans, 0, sizeof(TransBl) );
  memset ( &light, 0, sizeof(LightBl) );
  trans.trbuf = NewUniformTransBlock ();
  light.lsbuf = NewUniformLightBlock ();
  SetupModelMatrix ( model_rot_axis, model_rot_angle );
  InitViewMatrix ();
  ConstructIcosahedronVAO ();
  InitLights ();
  ResizeMyWorld ( width, height );
} /*InitMyWorld*/

void ResizeMyWorld ( int width, int height )
{
  float lr;

  glViewport ( 0, 0, win_width = width, win_height = height );
  SetupTextFrame ( width, height );
  lr = 0.5533*(float)width/(float)height;  /* przyjmujemy aspekt rowny 1 */
  M4x4Frustumf ( trans.pm, NULL, left = -lr, right = lr,
                 bottom = -0.5533, top = 0.5533, near = 5.0, far = 15.0 );
  LoadVPMatrix ( &trans );
  NotifyViewerPos ();
  ExitIfGLError ( "ResizeMyWorld" );
} /*ResizeMyWorld*/

void RedrawMyWorld ( char showpos )
{
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glEnable ( GL_DEPTH_TEST );
  if ( tesselate )
    DrawTessIcos ( option, enlight );
  else
    DrawIcosahedron ( option, enlight );
  if ( showpos )
    DisplayTextObject ( vptext );
  glFlush ();
  ExitIfGLError ( "RedrawMyWorld" );
} /*RedrawMyWorld*/

void DeleteMyWorld ( void )
{
  int i;

  glUseProgram ( 0 );
  for ( i = 0; i < 5; i++ )
    glDeleteProgram ( program_id[i] );
  glDeleteBuffers ( 1, &trans.trbuf );
  glDeleteBuffers ( 1, &light.lsbuf );
  glDeleteVertexArrays ( 1, &icos_vao );
  glDeleteBuffers ( 3, icos_vbo );
  DeleteTextObject ( vptext );
  for ( i = 0; i < 2; i++ )
    DeleteFontObject ( fonts[i] );
  TextCleanup ();
  DeleteEmptyVAO ();
#ifdef DEBUG_BUFFERS_ALLOCATION
  DumpBufferIdentifiers ();
#endif
  ExitIfGLError ( "DeleteMyWorld" );
} /*DeleteMyWorld*/

char ProcessCharCommand ( char charcode )
{
  int oldoption;

  oldoption = option;
  switch ( toupper ( charcode ) ) {
case 'L': enlight = !enlight;
    return true;
case 'M':
    font = font == fonts[0] ? fonts[1] : fonts[0];
    NotifyViewerPos ();
    return true;
case 'T':
    tesselate = !tesselate;
    return true;
case 'W': option = 0;  break;  /* przelaczamy na wierzcholki */
case 'K': option = 1;  break;  /* przelaczamy na krawedzie */
case 'S': option = 2;  break;  /* przelaczamy na sciany */
 default:
    return false;              /* ignorujemy wszystkie inne klawisze */
  }
  return option != oldoption;
} /*ProcessCharCommand*/

char MoveOn ( void )
{
  if ( (model_rot_angle += ANGULAR_VELOCITY * TimerTocTic ()) >= PI )
    model_rot_angle -= 2.0*PI;
  SetupModelMatrix ( model_rot_axis, model_rot_angle );
  return true;
} /*MoveOn*/

