#version 450 core

#define MAX_NLIGHTS 8

in FVertex {
    flat int instance;
    vec3     Colour;
    vec3     Position;
    vec3     pu, pv, Normal, TNormal;
    vec2     PatchCoord, TexCoord;
  } In;

out vec4 out_Colour;

uniform TransBlock {
    mat4 mm, mmti, vm, pm, vpm;
    vec4 eyepos;
  } trb;

struct LSPar {
    vec4 position;
    vec3 ambient;
    vec3 direct;
    vec3 attenuation;
  };

struct HLPar {
    vec3 zenith;
    vec4 sky, ground;
  };

uniform LSBlock {
    uint  nls;              /* liczba zrodel swiatla */
    uint  mask;             /* maska wlaczonych zrodel */
    bool  hlen;
    HLPar hl;
    LSPar ls[MAX_NLIGHTS];  /* poszczegolne zrola swiatla */
  } light;

struct Material {
    vec3  ambref;     /* odbijanie swiatla rozproszonego */
    vec3  dirref;     /* odbijanie swiatla kierunkowego */
    vec3  specref;    /* odbijanie zwierciadlane swiatla */
    float shininess, wa, we;  /* polysk */
  };

uniform MatBlock {
    Material mat;
  } mat;

uniform vec3 anc = vec3 ( 0.0, 1.0, 0.2 );

uniform bool NewtonIter = false, ModifyDepth = false;
uniform int ColourSource = 0, LightingModel = 0, NormalSource = 0;

layout (binding = 0) uniform sampler2D tex;

Material mm;
vec3     normal, tnormal, pnormal;

void GetColours ( vec2 tc )
{
  switch ( ColourSource ) {
default:
    mm.ambref = mm.dirref = In.Colour;
    mm.specref = vec3 ( 0.25 );
    mm.shininess = 20.0;  mm.wa = mm.we = 1.0;
    break;
case 1:
    mm = mat.mat;
    break;
case 2:
    mm = mat.mat;
    if ( tc.x >= 0.0 && tc.x <= 1.0 && tc.y >= 0.0 && tc.y <= 1.0 )
      mm.ambref = mm.dirref = texture ( tex, tc ).rgb;
    break;
  }
} /*GetColours*/

#define SQRT3 1.7320508
#define C (-0.005)
#define D -0.01
#define A 0.05
#define B (-0.003)

const mat3x2 dtxm1 = mat3x2 ( 10.0, 0.0,  0.0, 10.0,  0.0, 0.0 );
const mat3x2 dtxm2 = mat3x2 ( 6.0*SQRT3, 0.0,  0.0, 10.0,  0.0, 0.0 );
const mat3x2 dtxm3 = mat3x2 ( 7.0, -7.0, 7.0, 7.0, 0.0, 0.0 );
#define M4 3
#define N4 8
const mat3x2 dtxm4 = mat3x2 ( M4*SQRT3, -N4, N4*SQRT3, M4, 0.0, 0.0 );
const mat3x2 dtxm5 = mat3x2 ( 5.0, 0.0,  0.0, 5.0, 0.0, 0.0 );

vec3 NTex12 ( float dx, float dy )
{
  float r;

  r = 2.0*sqrt ( dx*dx + dy*dy );
  if ( r >= 1.0 )
    return vec3 ( 0.0, 0.0, 0.0 );
  else if ( r > 0.0 )
    return C*vec3 ( (r+r-3.0)*r*r+1.0, (24.0*r-24.0)*vec2 ( dx, dy ) );
  else
    return vec3 ( C, 0.0, 0.0 );
} /*NTex12*/

vec3 NormalTex12 ( float dx, float dy, mat2 L, out float dz )
{
  vec2 Ddq;
  vec3 d, dpu, dpv;

  d = NTex12 ( dx, dy );
  dz = d.x;
  Ddq = L * d.yz;
  dpu = In.pu + Ddq.x*pnormal;
  dpv = In.pv + Ddq.y*pnormal;
  return normalize ( cross ( dpu, dpv ) );
} /*NormalTex12*/

vec3 NormalTexture1 ( vec2 pc, out float dz )
{
  pc = fract ( dtxm1*vec3 ( pc, 1.0 ) );
  return NormalTex12 ( pc.x-0.5, pc.y-0.5, mat2(dtxm1), dz );
} /*NormalTexture1*/

vec3 NormalTexture2 ( vec2 pc, out float dz )
{
  int  c;
  float dx, dy;

  pc = dtxm2*vec3 ( pc, 1.0 );
  pc = fract ( vec2 ( pc.x/SQRT3, pc.y ) );
  switch ( (pc.x > 0.5 ? 1 : 0) + (pc.y > 0.5 ? 2 : 0) ) {
case 0: c = 3.0*pc.x-pc.y > 0.5 ? 1 : 0;  break;
case 1: c = 3.0*pc.x+pc.y > 2.5 ? 3 : 1;  break;
case 2: c = 3.0*pc.x+pc.y > 1.5 ? 2 : 0;  break;
case 3: c = 3.0*pc.x-pc.y > 1.5 ? 3 : 2;  break;
  }
  switch ( c ) {
case 0: dx = pc.x;      dy = pc.y-0.5;  break;  /* a */
case 1: dx = pc.x-0.5;  dy = pc.y;      break;  /* c */
case 2: dx = pc.x-0.5;  dy = pc.y-1.0;  break;  /* b */
case 3: dx = pc.x-1.0;  dy = pc.y-0.5;  break;  /* d */
  }
  return NormalTex12 ( dx*SQRT3, dy, mat2(dtxm2), dz );
} /*NormalTexture2*/

vec3 NTex34 ( float dx, float dy )
{
  float r;

  r = dx*dx + dy*dy;
  if ( r == 0.0 )
    return vec3 ( D, 0.0, 0.0 );
  else
    return D*vec3 ( 1.0-r, 2.0*dx, 2.0*dy );
} /*NTex34*/

vec3 NormalTex34 ( float dx, float dy, mat2 L, out float dz )
{
  vec2  Ddq;
  vec3  d, dpu, dpv;

  d = NTex34 ( dx, dy );
  dz = d.x;
  Ddq = L * d.yz;
  dpu = In.pu + Ddq.x*pnormal;
  dpv = In.pv + Ddq.y*pnormal;
  return normalize ( cross ( dpu, dpv ) );
} /*NormalTex34*/

vec3 NormalTexture3 ( vec2 pc, out float dz )
{
  pc = fract ( dtxm3*vec3 ( pc, 1.0 ) );
  return NormalTex34 ( pc.x-0.5, pc.y-0.5, mat2(dtxm3), dz );
} /*NormalTexture3*/

vec3 NormalTexture4 ( vec2 pc, out float dz )
{
  int  c;
  float dx, dy;

  pc = dtxm4*vec3 ( pc, 1.0 );
  pc = fract ( vec2 ( pc.x/SQRT3, pc.y ) );
  switch ( (pc.x > 0.5 ? 1 : 0) + (pc.y > 0.5 ? 2 : 0) ) {
case 0: c = 3.0*pc.x-pc.y > 0.5 ? 1 : 0;  break;
case 1: c = 3.0*pc.x+pc.y > 2.5 ? 3 : 1;  break;
case 2: c = 3.0*pc.x+pc.y > 1.5 ? 2 : 0;  break;
case 3: c = 3.0*pc.x-pc.y > 1.5 ? 3 : 2;  break;
  }
  switch ( c ) {
case 0: dx = pc.x;      dy = pc.y-0.5;  break;  /* a */
case 1: dx = pc.x-0.5;  dy = pc.y;      break;  /* c */
case 2: dx = pc.x-0.5;  dy = pc.y-1.0;  break;  /* b */
case 3: dx = pc.x-1.0;  dy = pc.y-0.5;  break;  /* d */
  }
  return NormalTex34 ( dx*SQRT3, dy, mat2(dtxm4), dz );
} /*NormalTexture4*/

vec3 NTex56 ( float dx, float dy )
{
  float x, d, ds, dt;

  if ( dx >= 0.5 ) {
    dx -= 0.5;
    dy = dy >= 0.5 ? dy-0.5 : dy+0.5;
  }
  ds = dt = 0.0;
  if ( dx <= A ) {
    if ( dy <= dx ) { x = dy/A;  ds = 2.0*x/A; }
    else if ( dy >= 1.0-dx ) { x = (dy-1.0)/A;  ds = 2.0*x/A; }
    else { x = dx/A;  dt = 2.0*x/A; }
  }
  else if ( dx >= 0.5-A ) {
    if ( dx <= 0.5-dy ) { x = dy/A;  ds = 2.0*x/A; }
    else if ( dx <= dy-0.5 ) { x = (dy-1.0)/A;  ds = 2.0*x/A; }
    else { x = (dx-0.5)/A;  dt = 2.0*x/A; }
  }
  else {
    if ( dy <= A ) { x = dy/A;  ds = 2.0*x/A; }
    else if ( dy >= 1.0-A ) { x = (dy-1.0)/A;  ds = 2.0*x/A; }
    else return vec3 ( B, 0.0, 0.0 );
  }
  d = x*x;
  return B*vec3 ( d, dt, ds );
} /*NTex56*/

vec3 NormalTex56 ( vec2 pc, mat2 L, out float dz )
{
  vec2 Ddq;
  vec3 d, dpu, dpv;

  d = NTex56 ( pc.x, pc.y );
  dz = d.x;
  Ddq = L * d.yz;
  dpu = In.pu + Ddq.x*pnormal;
  dpv = In.pv + Ddq.y*pnormal;
  return normalize ( cross ( dpu, dpv ) );
} /*NormalTex56*/

vec3 NormalTexture5 ( vec2 pc, out float dz )
{
  if ( In.instance < 4 || In.instance >= 8 )
    { dz = 0.0;  return pnormal; }
  pc = fract ( dtxm5*vec3 ( pc, 1.0 ) );
  return NormalTex56 ( pc, mat2(dtxm5), dz );
} /*NormalTexture5*/

vec3 NormalTexture6 ( vec2 pc, out float dz )
{
  vec2  Ddq, f, pd;
  vec3  dd, dpu, dpv, APu, APv, AN;
  mat2  Df;

  if ( In.instance < 4 || In.instance >= 8 )
    { dz = 0.0;  return pnormal; }
  pd = fract ( dtxm5*vec3 ( pc, 1.0 ) );
  dd = NTex56 ( pd.x, pd.y );
  APu = mat3(trb.vpm) * In.pu;
  APv = mat3(trb.vpm) * In.pv;
  AN = mat3(trb.vpm) * pnormal;
  f = dd.x*AN.xy;
  Df = mat2 ( APu.xy, APv.xy );
  pc -= inverse ( Df )*f;
  pd = fract ( dtxm5*vec3 ( pc, 1.0 ) );
  dd = NTex56 ( pd.x, pd.y );
  dz = dd.x;
  Ddq = mat2(dtxm5) * dd.yz;
  dpu = In.pu + Ddq.x*pnormal;
  dpv = In.pv + Ddq.y*pnormal;
  return normalize ( cross ( dpu, dpv ) );
} /*NormalTexture6*/

#define M3ROW(M,I) vec3 ( M[0][I], M[1][I], M[2][I] )

vec3 GetNormalVector ( vec2 pc, out float dz )
{
  vec3  nv;

  switch ( NormalSource ) {
default: dz = 0.0; return pnormal;
 case 1: nv = NormalTexture1 ( pc, dz );  break;
 case 2: nv = NormalTexture2 ( pc, dz );  break;
 case 3: nv = NormalTexture3 ( pc, dz );  break;
 case 4: nv = NormalTexture4 ( pc, dz );  break;
 case 5: nv = NormalTexture5 ( pc, dz );  break;
 case 6: nv = NormalTexture6 ( pc, dz );  break;
  }
  dz *= 0.5*dot ( M3ROW ( trb.vpm, 2 ), pnormal ) * gl_FragCoord.w;
  return nv;
} /*GetNormalVector*/

vec3 HemisphericalLight ( vec3 normal, float e )
{
  float c;

  c = dot ( normal, light.hl.zenith );
  return mm.dirref*mix ( light.hl.sky.rgb, light.hl.ground.rgb,
                         e > 0.0 ? 0.5-0.5*c : 0.5+0.5*c );
} /*HemisphericalLight*/

vec3 posDifference ( vec4 p, vec3 pos, out float dist )
{
  vec3 v;

  if ( p.w != 0.0 ) {
    v = p.xyz/p.w-pos.xyz;
    dist = sqrt ( dot ( v, v ) );
  }
  else
    v = p.xyz;
  return normalize ( v );
} /*posDifference*/

float attFactor ( vec3 att, float dist )
{
  return 1.0/(((att.z*dist)+att.y)*dist+att.x);
} /*attFactor*/

vec3 LambertLighting ( void )
{
  vec3  lv, vv, Colour;
  float d, e, dist;
  uint  i, mask;

  if ( light.hlen )
    Colour = HemisphericalLight ( normal, e );
  else
    Colour = vec3 ( 0.0 );
  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, tnormal );
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      Colour += light.ls[i].ambient * mm.ambref;
      lv = posDifference ( light.ls[i].position, In.Position, dist );
      d = dot ( lv, normal );
      if ( e > 0.0 ) {
        if ( d > 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            d *= attFactor ( light.ls[i].attenuation, dist );
          Colour += (d * light.ls[i].direct) * mm.dirref;
        }
      }
      else {
        if ( d < 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            d *= attFactor ( light.ls[i].attenuation, dist );
          Colour -= (d * light.ls[i].direct) * mm.dirref;
        }
      }
    }
  return clamp ( Colour, 0.0, 1.0 );
} /*LambertLighting*/

float wFactor ( float lvn, float wa, float we )
{
  return 1.0+(wa-1.0)*pow ( 1.0-lvn, we );
} /*wFactor*/

vec3 BlinnPhongLighting ( void )
{
  vec3  lv, vv, hv, Colour;
  float a, d, e, f, dist;
  uint  i, mask;

  if ( light.hlen )
    Colour = HemisphericalLight ( normal, e );
  else
    Colour = vec3 ( 0.0 );
  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, tnormal );
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      Colour += light.ls[i].ambient * mm.ambref;
      lv = posDifference ( light.ls[i].position, In.Position, dist );
      d = dot ( lv, normal );
      if ( e > 0.0 ) {
        if ( d > 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            a = attFactor ( light.ls[i].attenuation, dist );
          else
            a = 1.0;
          Colour += (a * d * light.ls[i].direct) * mm.dirref;
          hv = normalize ( lv+vv );
          f = pow ( dot ( hv, normal ), mm.shininess );
          Colour += (a*f*wFactor(d,mm.wa,mm.we))*mm.specref;
        }
      }
      else {
        if ( d < 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            a = attFactor ( light.ls[i].attenuation, dist );
          else
            a = 1.0;
          Colour -= (a * d * light.ls[i].direct) * mm.dirref;
          hv = normalize ( lv+vv );
          f = pow ( -dot ( hv, normal ), mm.shininess );
          Colour += (a*f*wFactor(-d,mm.wa,mm.we))*mm.specref;
        }
      }
    }
  return clamp ( Colour, 0.0, 1.0 );
} /*BlinnPhongLighting*/

vec3 AnisotropicHVector ( vec3 w, vec3 normal, vec3 rv )
{
  vec3 p0w, p1w;

  p0w = normal*dot ( normal, w );
  p1w = rv*dot ( rv, w );
  return normalize ( mix ( p0w+p1w, w, anc.z ) );
} /*AnisotroicHVector*/

vec3 AnisotropicLighting ( void )
{
  vec3  rv, lv, vv, hv, Colour;
  float a, d, e, f, dist;
  uint  i, mask;

  if ( light.hlen )
    Colour = HemisphericalLight ( normal, e );
  else
    Colour = vec3 ( 0.0 );
  rv = anc.x*In.pu + anc.y*In.pv;
  rv = normalize ( rv - normal*dot ( rv, normal ) );
  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, tnormal );
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      Colour += light.ls[i].ambient * mm.ambref;
      lv = posDifference ( light.ls[i].position, In.Position, dist );
      d = dot ( lv, normal );
      if ( e > 0.0 ) {
        if ( d > 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            a = attFactor ( light.ls[i].attenuation, dist );
          else
            a = 1.0;
          Colour += (a * d * light.ls[i].direct) * mm.dirref;
          hv = AnisotropicHVector ( lv+vv, normal, rv  );
          f = pow ( dot ( hv, normal ), mm.shininess );
          Colour += 0.65*(a*f*wFactor(d,mm.wa,mm.we))*mm.specref;
        }
      }
      else {
        if ( d < 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            a = attFactor ( light.ls[i].attenuation, dist );
          else
            a = 1.0;
          Colour -= (a * d * light.ls[i].direct) * mm.dirref;
          hv = AnisotropicHVector ( lv+vv, normal, rv );
          f = pow ( -dot ( hv, normal ), mm.shininess );
          Colour += 0.65*(a*f*wFactor(-d,mm.wa,mm.we))*mm.specref;
        }
      }
    }
  return clamp ( Colour, 0.0, 1.0 );
} /*AnisotropicLighting*/

#define AGamma(colour) pow ( colour, vec3(256.0/563.0) )

void main ( void )
{
  vec3  Colour;
  float dz;

  pnormal = normalize ( In.Normal );
  tnormal = In.TNormal;
  normal = GetNormalVector ( In.PatchCoord, dz );
  gl_FragDepth = ModifyDepth ? gl_FragCoord.z + dz : gl_FragCoord.z;
  GetColours ( In.TexCoord );
  switch ( LightingModel ) {
default: Colour = LambertLighting ();      break;
 case 1: Colour = BlinnPhongLighting ();   break;
 case 2: Colour = AnisotropicLighting ();  break;
  }
  out_Colour = vec4 ( AGamma ( Colour ), 1.0 );
} /*main*/
