#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "../utilities/openglheader.h"

#include "../utilities/utilities.h"
#include "trans.h"

#define NTRUOFFS 6

static GLuint trbbp = GL_INVALID_INDEX;
static GLint  trbsize, trbofs[NTRUOFFS];
static const GLchar *UTBNames[] =
  { "TransBlock", "TransBlock.mm", "TransBlock.mmti", "TransBlock.vm",
    "TransBlock.pm", "TransBlock.vpm", "TransBlock.eyepos" };

GLuint GetAccessToTransBlockUniform ( GLuint program_id )
{
  if ( trbbp == GL_INVALID_INDEX )
    GetAccessToUniformBlock ( program_id, NTRUOFFS, &UTBNames[0],
                              &trbsize, trbofs, &trbbp );
  else
    AttachUniformBlockToBP ( program_id, UTBNames[0], trbbp );
  return trbbp;
} /*GetAccessToTransBlockUniform*/

GLuint NewUniformTransBlock ( void )
{
  return NewUniformBuffer ( trbsize, trbbp );
} /*NewUniformTransBlock*/

void AttachUniformTransBlockToBP ( GLuint program_id )
{
  AttachUniformBlockToBP ( program_id, UTBNames[0], trbbp );
} /*AttachUniformTransBlockToBP*/

void LoadVPMatrix ( TransBl *trans, char mirror )
{
  GLfloat *vm, *pm, *ep, vpm[16];

  if ( mirror )
    { vm = trans->mvm;  pm = trans->mpm;  ep = trans->reyepos; }
  else
    { vm = trans->wvm;  pm = trans->wpm;  ep = trans->eyepos; }
  M4x4Multf ( vpm, pm, vm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trans->trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 16*sizeof(GLfloat), pm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 16*sizeof(GLfloat), vpm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[5], 4*sizeof(GLfloat), ep );
  ExitIfGLError ( "LoadVPMatrix" );
} /*LoadPMatrix*/

void LoadMMatrix ( TransBl *trans, GLfloat mm[16] )
{
  GLfloat mmti[16];

  if ( mm )
    memcpy ( trans->mm, mm, 16*sizeof(GLfloat) );
  glBindBuffer ( GL_UNIFORM_BUFFER, trans->trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[0], 16*sizeof(GLfloat), trans->mm );
  M4x4TInvertf ( mmti, trans->mm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[1], 16*sizeof(GLfloat), mmti );
  ExitIfGLError ( "LoadMMatrix" );
} /*LoadMMatrix*/

void LoadShTrans ( TransBl *trans, GLfloat vm[16], GLfloat pm[16], GLfloat eyepos[4] )
{
  GLfloat vpm[16];

  M4x4Multf ( vpm, pm, vm );
  glBindBuffer ( GL_UNIFORM_BUFFER, trans->trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[2], 16*sizeof(GLfloat), vm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[3], 16*sizeof(GLfloat), pm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[4], 16*sizeof(GLfloat), vpm );
  glBufferSubData ( GL_UNIFORM_BUFFER, trbofs[5], 4*sizeof(GLfloat), eyepos );
  ExitIfGLError ( "LoadShTrans" );
} /*LoadShTrans*/

