#version 450 core

#define MAX_DEG 10

#define MIN_TESS_PARAM  3.0
#define MAX_TESS_PARAM 32.0

layout(local_size_x=1) in;

layout(std430,binding=0) buffer CPoints {
    float cp[];
  } cp;

layout(std430,binding=1) buffer CPIndices {
    int cpi[];
  } cpi;

layout(std430,binding=2) buffer BezPatch {
    int  npatches, dim, udeg, vdeg;
    int  stride_u, stride_v, stride_p, stride_q, nq;
    bool use_ind;
    vec3 Colour;
    int  TessLevel;
    bool BezNormals;
  } bezp;

layout(std430,binding=4) buffer BezPatchTessParams { float tsp[]; };

uniform TransBlock {
    mat4  mm, mmti, vm, pm, vpm;
    vec4  eyepos;
    float left, right, bottom, top, near, far;
    vec4  viewport;
  } trb;

uniform float maxangle = 15.0*0.017453292519943278;  /* 15 stopni */

int  ncp;
vec4 ctp[(MAX_DEG+1)*(MAX_DEG+1)], bcp[MAX_DEG+1];

vec4 BCHorner4f ( int n, float t )
{
  int   i, b;
  float s, d;
  vec4  q;

  s = 1.0-t;  d = t;  b = n;
  q = bcp[0];
  for ( i = 1; i <= n; i++ ) {
    q = s*q + (float(b)*d)*bcp[i];
    d *= t;  b = (b*(n-i))/(i+1);
  }
  return q;
} /*BCHorner4f*/

vec3 BCHornerR3f ( int n, float t )
{
  vec4 q;

  q = BCHorner4f ( n, t );
  return q.xyz / q.w;
} /*BCHornerR3f*/

int AngBCDens ( int n )
{
  int   d, i;
  float t, l1, l2, ang;
  vec3  v1, v2, p1, p2;
  bool  z;

  for ( d = max ( n, int(MIN_TESS_PARAM) ); d <= int(MAX_TESS_PARAM); d++ ) {
    p1 = BCHornerR3f ( n, 1.0/float(d) );
    v1 = p1 - bcp[0].xyz / bcp[0].w;  z = (l1 = length ( v1 )) == 0.0;
    for ( i = 2, ang = 0.0;  i <= d;  i++ ) {
      p2 = BCHornerR3f ( n, float(i)/float(d) );
      v2 = p2 - p1;  l2 = length ( v2 );
      if ( l2 > 0.0 ) {
        z = false;
        if ( l1 > 0.0 && (ang = acos ( dot ( v1, v2 ) / (l1*l2) )) > maxangle )
          break;
        p1 = p2;  v1 = v2;  l1 = l2;
      }
    }
    if ( z || ang <= maxangle )
      break;
  }
  return d;
} /*AngBCDens*/

void FindShapeParams ( int l )
{
  int   i, j, k, m, d, e;
  float t;
  vec4  bc[MAX_DEG+1];

  for ( i = e = 0;  i <= bezp.udeg;  i++ ) {
    t = float(i)/float(bezp.udeg);
    for ( j = 0; j <= bezp.vdeg; j++ ) {
      for ( m = 0, k = j;  m <= bezp.udeg;  m++, k += bezp.vdeg+1 )
        bcp[m] = ctp[k];
      bc[j] = BCHorner4f ( bezp.udeg, t );
    }
    for ( j = 0; j <= bezp.vdeg; j++ ) bcp[j] = bc[j];
    if ( (d = AngBCDens ( bezp.vdeg )) > e )
      e = d;
    if ( i == 0 ) tsp[l+3] = float(d);
  }
  tsp[l+4] = float(e);
  tsp[l+5] = float(d);
  for ( j = e = 0;  j <= bezp.vdeg;  j++ ) {
    t = float(j)/float(bezp.vdeg);
    for ( i = k = 0;  i <= bezp.udeg;  i++ ) {
      for ( m = 0;  m <= bezp.vdeg;  m++, k++ )
        bcp[m] = ctp[k];
      bc[i] = BCHorner4f ( bezp.vdeg, t );
    }
    for ( i = 0; i <= bezp.udeg; i++ ) bcp[i] = bc[i];
    if ( (d = AngBCDens ( bezp.udeg )) > e )
      e = d;
    if ( j == 0 ) tsp[l] = float(d);
  }
  tsp[l+1] = float(e);
  tsp[l+2] = float(d);
} /*FindShapeParams*/

void main ( void )
{
  int inst, i, j, k, l, i0;
#define FINDCPOINTS(DIM,Z,W) { \
    if ( bezp.use_ind ) { \
      i0 = inst*bezp.stride_p; \
      for ( i = k = 0;  i <= bezp.udeg;  i++ ) { \
        for ( j = 0;  j <= bezp.vdeg;  j++, k++ ) { \
          l = DIM*cpi.cpi[i0+k]; \
          ctp[k] = trb.mm * vec4 ( cp.cp[l], cp.cp[l+1], Z, W ); \
    } } } \
    else { \
      i0 = (inst / bezp.nq)*bezp.stride_p+(inst % bezp.nq)*bezp.stride_q; \
      for ( i = k = 0;  i <= bezp.udeg;  i++ ) { \
        for ( j = 0, l = i0+i*bezp.stride_u;  j <= bezp.vdeg; \
              j++, k++, l += bezp.stride_v ) { \
          ctp[k] = trb.mm * vec4 ( cp.cp[l], cp.cp[l+1], Z, W ); \
    } } } }

  inst = int(gl_GlobalInvocationID.x);
  ncp = (bezp.udeg+1)*(bezp.vdeg+1);
  switch ( bezp.dim ) {
case 2: FINDCPOINTS ( 2, 0.0, 1.0 )  break;
case 3: FINDCPOINTS ( 3, cp.cp[l+2], 1.0 )  break;
case 4: FINDCPOINTS ( 4, cp.cp[l+2], cp.cp[l+3] )  break;
  }
  FindShapeParams ( 12*inst+6 );
} /*main*/
