#version 450 core

layout (local_size_x=1) in;

layout (std430,binding = 0) buffer RCA    { uint  rc[]; } rca;
layout (std430,binding = 1) buffer CoeffA { float a[]; }  aa;
layout (std430,binding = 2) buffer RCB    { uint  rc[]; } rcb;
layout (std430,binding = 3) buffer CoeffB { float a[]; }  ab;
layout (std430,binding = 4) buffer Auxb0  { uint  a[]; }  seq;
layout (std430,binding = 5) buffer Auxb1  { float a[]; }  aux1;
layout (std430,binding = 6) buffer Auxb2  { uint  a[]; }  aux2;

uniform int  stage;
uniform uint prN0, prN, prStep;
uniform uint ma, nnza, mb, nprod, nnzc, h, tablgt;
uniform bool reverse;

#define ra(I) rca.rc[I]
#define ca(I) rca.rc[ma+1+(I)]
#define rb(I) rcb.rc[I]
#define cb(I) rcb.rc[mb+1+(I)]
#define pairi(I) aux2.a[2*(I)]
#define pairj(I) aux2.a[2*(I)+1]
#define tab1(I) seq.a[I]
#define tab2(I) seq.a[tablgt+(I)]

void iPrefixSum ( uint i )
{
  uint ii, m0, m1, ia, ib;

  ii = i+i;  m0 = 0x01 << prStep;  m1 = m0-1;
  ia = (ii & ~m0) | m1;
  if ( (ib = ia + (i & m1) + 1) < prN )
    seq.a[prN0 + ib] += seq.a[prN0 + ia];
} /*iPrefixSum*/

void SortIt ( uint ns, uint np )
{
  uint  i, j, l, h2;
  float x;

  h2 = h >> 1;  l = np % h2;  np /= h2;
  i = tab1(ns)+np*h+l;  j = reverse ? tab1(ns)+(np+1)*h-l-1 : i + h2;
  if ( j < tab1(ns+1) ) {
    if ( pairj(i) > pairj(j) ) {
      l = pairj(i);   pairj(i) = pairj(j);    pairj(j) = l;
      x = aux1.a[i];  aux1.a[i] = aux1.a[j];  aux1.a[j] = x;
    }
  }
} /*SortIt*/

void main ( void )
{
  uint  i, j, k, l, m, p, q;
  float s;

  i = gl_GlobalInvocationID.x;
  switch ( stage ) {
case 0:
    iPrefixSum ( i );
    return;
case 1:
    if ( i == 0 ) seq.a[i] = 0;
    j = ca(i);
    seq.a[i+1] = rb(j+1)-rb(j);
    return;
case 2:
    for ( p = 0, k = nnza;  k-p > 1; ) {
      l = p + (k-p)/2;
      if ( i >= seq.a[l] ) p = l; else k = l;
    }
    for ( j = 0, k = ma;  k-j > 1; ) {
      l = j + (k-j)/2;
      if ( p >= ra(l) ) j = l; else k = l;
    }
    pairi(i) = j;
    q = rb(ca(p))+i-seq.a[p];
    pairj(i) = cb(q);
    aux1.a[i] = aa.a[p]*ab.a[q];
    return;
case 3:
    if ( i == 0 ) {
      tab1(0) = 0;
      tab1(ma) = nprod;
    }
    else {
      for ( j = 0, k = nprod;  k-j > 1; ) {
        l = j + (k-j)/2;
        if ( pairi(l) < i ) j = l; else k = l;
      }
      tab1(i) = k;
    }
    return;
case 4:
    tab2(i) = tab1(i+1)-tab1(i);
    return;
case 5:
    if ( (j = i+(prN+1)/2) < prN ) {
      if ( tab2(i) < tab2(j) )
        tab2(i) = tab2(j);
    }
    return;
case 6:
    SortIt ( i, gl_GlobalInvocationID.y );
    return;
case 7:
    if ( i == 0 )
      { tab1(0) = 0;  tab1(1) = 1; }
    else
      tab1(i+1) = pairi(i-1) != pairi(i) || pairj(i-1) != pairj(i) ? 1 : 0;
    return;
case 8:
    if ( i == nprod )
      tab2(nnzc) = nprod;
    else if ( tab1(i+1) > tab1(i) ) {
      ca(tab1(i)) = pairj(i);
      tab2(tab1(i)) = i;
    }
    return;
case 9:
    if ( tab2(i+1) > tab2(i) ) {
      for ( j = tab2(i), s = 0.0;  j < tab2(i+1);  j++ )
        s += aux1.a[j];
      aa.a[i] = s;
      tab1(i) = pairi(tab2(i));
    }
    return;
case 10:
    if ( i == 0 ) {
      ra(0) = 0;
      ra(ma) = nnzc;
    }
    else {
      for ( p = 0, k = nnzc;  k-p > 1; ) {
        l = p + (k-p)/2;
        if ( i > tab1(l-1) ) p = l; else k = l;
      }
      ra(i) = p;
    }
    return;
default:
    return;
  }
} /*main*/

