#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <sys/times.h> 
#include <math.h>
#include "openglheader.h"

#include "utilities.h"
#include "trans.h"
#include "lights.h"
#include "bspatches.h"
#include "app2c.h"

GLuint progid[2];

GLuint ubsnloc, ubstloc;
GLint  ColourSourceLoc, LightingModelLoc;

BSPatchObjf *mypatch;

float   model_rot_axis[3] = {0.0,0.0,1.0},
        patch_rot_angle0 = 0.0, patch_rot_angle = 0.0;
GLfloat patch_mmatrix[16], patch_mmti[16];
TransBl trans;
LightBl light;
GLuint  matbuf[2];
GLint   lighting_model = 0;
char    cnet = 0, skeleton = 0;


void LoadMyShaders ( void )
{
  static const char *filename[] =
    { "app2.vert.glsl", "bspatch.tesc.glsl", "bspatch.tese.glsl",
      "app2.geom.glsl", "app2c.frag.glsl", "app1.vert.glsl", "app1.frag.glsl" };
  static const GLuint shtype[9] =
    { GL_VERTEX_SHADER, GL_TESS_CONTROL_SHADER, GL_TESS_EVALUATION_SHADER,
      GL_GEOMETRY_SHADER, GL_FRAGMENT_SHADER,
      GL_VERTEX_SHADER, GL_FRAGMENT_SHADER };
  static const GLchar *UVarNames[] = { "ColourSource", "LightingModel" };

  GLuint shader_id[7];
  GLint i;

  for ( i = 0; i < 7; i++ )
    shader_id[i] = CompileShaderFiles ( shtype[i], 1, &filename[i] );
  progid[0] = LinkShaderProgram ( 5, shader_id, "0" );
  progid[1] = LinkShaderProgram ( 2, &shader_id[5], "1" );
/*PrintProgramResources ( progid[0], "0" );*/
  GetAccessToTransBlockUniform ( progid[0] );
  GetAccessToLightMatUniformBlocks ( progid[0] );
  GetAccessToBSPatchStorageBlocks ( progid[0] );
  ColourSourceLoc = glGetUniformLocation ( progid[0], UVarNames[0] );
  LightingModelLoc = glGetUniformLocation ( progid[0], UVarNames[1] );
  trans.trbuf = NewUniformTransBlock ();
  light.lsbuf = NewUniformLightBlock ();
  matbuf[1] = NewUniformMatBlock ();
  matbuf[0] = NewUniformMatBlock ();
  AttachUniformTransBlockToBP ( progid[1] );
  for ( i = 0; i < 7; i++ )
    glDeleteShader ( shader_id[7] );
  ExitIfGLError ( "LoadMyShaders" );
} /*LoadMyShaders*/

void SetupMVPMatrix ( void )
{
  GLfloat m[16], mvp[16];

  M4x4Multf ( m, trans.vm, trans.mm );
  M4x4Multf ( mvp, trans.pm, m );
  LoadVPMatrix ( &trans );
  LoadMMatrix ( &trans, NULL );
} /*SetupMVPMatrix*/

void SetModelMatrix ( GLfloat mm[16], GLfloat mmti[16] )
{
  memcpy ( trans.mm, mm, 16*sizeof(GLfloat) );
  LoadMMatrix ( &trans, NULL );
} /*SetModelMatrix*/

void SetupPatchMatrix ( void )
{
  M4x4RotateVf ( patch_mmatrix, model_rot_axis[0], model_rot_axis[1], model_rot_axis[2],
                 animate ? patch_rot_angle : patch_rot_angle0 );
  M4x4TInvertf ( patch_mmti, patch_mmatrix );
} /*SetupTeapotMatrix*/

void InitViewMatrix ( void )
{
  memcpy ( trans.eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  M4x4Translatef ( trans.vm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  LoadVPMatrix ( &trans );
} /*InitViewMatrix*/

void RotateViewer ( double delta_xi, double delta_eta )
{
  float   vi[3], lgt, vk[3];
  double  angi, angk;
  GLfloat tm[16], rm[16];

  if ( delta_xi == 0 && delta_eta == 0 )
    return;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  vi[0] = (float)delta_eta*(right-left)/(float)win_height;
  vi[1] = (float)delta_xi*(top-bottom)/(float)win_width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;
  vi[1] /= lgt;
  angi = -0.052359878;  /* -3 stopnie */
  V3CompRotationsf ( vk, &angk, viewer_rvec, viewer_rangle, vi, angi );
  memcpy ( viewer_rvec, vk, 3*sizeof(float) );
  viewer_rangle = angk;
  M4x4Translatef ( tm, -viewer_pos0[0], -viewer_pos0[1], -viewer_pos0[2] );
  M4x4RotateVf ( rm, viewer_rvec[0], viewer_rvec[1], viewer_rvec[2],
                 -viewer_rangle );
  M4x4Multf ( trans.vm, tm, rm );
  M4x4Transposef ( tm, rm );
  M4x4MultMVf ( trans.eyepos, tm, viewer_pos0 );
  LoadVPMatrix ( &trans );
} /*RotateViewer*/

void ConstructMyPatch ( void )
{
  const GLfloat MyColour[4] = { 0.95, 0.95, 0.2, 1.0 };
  const GLfloat ambr[4]  = { 0.75, 0.6, 0.2, 1.0 };
  const GLfloat diffr[4] = { 0.75, 0.6, 0.2, 1.0 };
  const GLfloat specr[4] = { 0.7, 0.7, 0.6, 1.0 };
  const GLfloat shn = 60.0, wa = 5.0, we = 5.0;
  const GLfloat knots_u[] = {-0.1, 0.0, 0.0, 0.0, 0.25, 0.5, 0.75, 1.0, 1.0, 1.0, 1.1};
  const GLfloat knots_v[] =  {-0.1, 0.0, 0.0, 0.0, 0.0, 0.375, 0.75, 1.0, 1.0, 1.0, 1.0, 1.1};
  const GLfloat cpoints[][3] = {
    {-1.005681565231,-0.921225247873,-0.391540306634}, {-1.005681565231,-0.772318640198,-0.335890887386},
    {-1.005681565231,-0.474057241266,-0.219388678784}, {-0.848832146801,-0.020862068966,-0.034215671371},
    {-0.732020383479,0.479137931034,-0.028688524457}, {-0.732020383479,0.820172413793,-0.028688524457},
    {-0.732020383479,1.0,-0.028688524457},
/*  % 1:*/
    {-0.852635772597,-1.181582354547,0.305122732483}, {-0.852635772597,-1.032675746872,0.360772151732},
    {-0.852635772597,-0.706671377557,0.403039774198}, {-0.719655881458,-0.020862068966,0.540607607664},
    {-0.620620668413,0.479137931034,0.453278686426}, {-0.620620668413,0.820172413793,0.453278686426},
    {-0.620620668413,1.0,0.453278686426},
/*  % 2:*/
    {-0.499270490297,-0.921225247873,-0.391540306634}, {-0.499270490297,-0.772318640198,-0.335890887386},
    {-0.499270490297,-0.474057241266,-0.219388678784}, {-0.421402615663,-0.020862068966,-0.034215671371},
    {-0.363411430022,0.479137931034,-0.028688524457}, {-0.363411430022,0.820172413793,-0.028688524457},
    {-0.363411430022,1.0,-0.028688524457},
/*  % 3:*/
    {0.016854377356,-1.181582354547,0.305122732483}, {0.016854377356,-1.032675746872,0.360772151732},
    {0.016854377356,-0.706671377557,0.403039774198}, {0.014225712998,-0.020862068966,0.540607607664},
    {0.012268046072,0.479137931034,0.453278686426}, {0.012268046072,0.820172413793,0.453278686426},
    {0.012268046072,1.0,0.453278686426},
/*  % 4:*/
    {0.530475251389,-0.921225247873,-0.391540306634}, {0.530475251389,-0.772318640198,-0.335890887386},
    {0.530475251389,-0.474057241266,-0.219388678784}, {0.447740579154,-0.020862068966,-0.034215671371},
    {0.386124903123,0.479137931034,-0.028688524457}, {0.386124903123,0.820172413793,-0.028688524457},
    {0.386124903123,1.0,-0.028688524457},
/*  % 5:*/
    {0.843776635755,-1.181582354547,0.305122732483}, {0.843776635755,-1.032675746872,0.360772151732},
    {0.843776635755,-0.706671377557,0.403039774198}, {0.712178444858,-0.020862068966,0.540607607664},
    {0.61417223685,0.479137931034,0.453278686426}, {0.61417223685,0.820172413793,0.453278686426},
    {0.61417223685,1.0,0.453278686426},
/*  % 6:*/
    {0.994318434769,-0.921225247873,-0.391540306634}, {0.994318434769,-0.772318640198,-0.335890887386},
    {0.994318434769,-0.474057241266,-0.219388678784}, {0.839241247695,-0.020862068966,-0.034215671371},
    {0.723749332874,0.479137931034,-0.028688524457}, {0.723749332874,0.820172413793,-0.028688524457},
    {0.723749332874,1.0,-0.028688524457}};

  mypatch = EnterBSplinePatch ( 3, 10, knots_u, 4, 11, knots_v,
                                3, 21, &cpoints[0][0], MyColour );
  SetBSPatchNormals ( mypatch, BSNormals );
  SetBSPatchTessLevel ( mypatch, TessLevel );
  SetupMaterial ( matbuf[0], ambr, diffr, specr, shn, wa, we );
} /*ConstructMyPatch*/

void InitMyObject ( void )
{
  struct tms clk;

  clocks_per_sec = (float)sysconf(_SC_CLK_TCK);
  zero_time = times ( &clk );
  memset ( &trans, 0, sizeof(TransBl) );
  memset ( &light, 0, sizeof(LightBl) );
  LoadMyShaders ();
  ConstructEmptyVAO ();
  SetupPatchMatrix ();
  InitViewMatrix ();
  ConstructMyPatch ();
  InitLights ();
  glUseProgram ( progid[0] );
  glUniform1i ( ColourSourceLoc, 0 );
} /*InitMyObject*/

void InitLights ( void )
{
  GLfloat amb0[3] = { 0.2, 0.2, 0.3 };
  GLfloat dir0[3] = { 0.8, 0.8, 0.8 };
  GLfloat pos0[4] = { 0.0, 1.0, 1.0, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  SetLightAmbient ( &light, 0, amb0 );
  SetLightDirect ( &light, 0, dir0 );
  SetLightPosition ( &light, 0, pos0 );
  SetLightAttenuation ( &light, 0, atn0 );
  SetLightOnOff ( &light, 0, true );
} /*InitLights*/

void DrawMyPatch ( void )
{
  if ( skeleton )
    glPolygonMode ( GL_FRONT_AND_BACK, GL_LINE );
  else
    glPolygonMode ( GL_FRONT_AND_BACK, GL_FILL );
  glUseProgram ( progid[0] );
  ChooseMaterial ( 0 );
  DrawBSplinePatch ( mypatch );
  glUseProgram ( progid[1] );
  DrawBSplineCPoints ( mypatch );
} /*DrawMyPatch*/

