#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h"

#include "utilities.h"
#include "bspatches.h"

#define NBSPLINEPATCHOFFS 9

static GLuint bspbbp = GL_INVALID_INDEX, bsknbbp = GL_INVALID_INDEX,
              bscpbbp = GL_INVALID_INDEX;
static GLint  bspbsize, bspbofs[NBSPLINEPATCHOFFS];

static const GLchar *BSKNNames[] = { "BSKnots" };
static const GLchar *BSCPNames[] = { "BSCPoints" };
static const GLchar *BSPNames[] =
    { "BSPatch", "BSPatch.dim", "BSPatch.n", "BSPatch.N", "BSPatch.m",
      "BSPatch.M", "BSPatch.stride", "BSPatch.Colour", "BSPatch.BSPNormals",
      "BSPatch.tesslevel" };

void GetAccessToBSPatchStorageBlocks ( GLuint program_id )
{
  GLint size, ofs;

  if ( bspbbp == GL_INVALID_INDEX )
    GetAccessToStorageBlock ( program_id, NBSPLINEPATCHOFFS, &BSPNames[0],
                              &bspbsize, bspbofs, &bspbbp );
  if ( bsknbbp == GL_INVALID_INDEX )
    GetAccessToStorageBlock ( program_id, 0, &BSKNNames[0],
                              &size, &ofs, &bsknbbp );
  if ( bscpbbp == GL_INVALID_INDEX )
    GetAccessToStorageBlock ( program_id, 0, &BSCPNames[0],
                              &size, &ofs, &bscpbbp );
} /*GetAccessToBSPatchStorageBlocks*/

BSPatchObjf *EnterBSplinePatch ( GLint udeg, GLint lknu, const GLfloat *knotsu,
                                 GLint vdeg, GLint lknv, const GLfloat *knotsv,
                                 GLint dim, GLint stride, const GLfloat *cp,
                                 const GLfloat *colour )
{
  BSPatchObjf *bsp;
  GLint one = GL_TRUE, ten = 10;

  if ( dim < 2 || dim > 4 || udeg < 1 || udeg > MAX_BSPATCH_DEG ||
       vdeg < 1 || vdeg > MAX_BSPATCH_DEG || lknu <= 2*udeg || lknv <= 2*vdeg )
    return NULL;
  bsp = malloc ( sizeof(BSPatchObjf) );
  if ( bsp ) {
    memset ( bsp, 0, sizeof(BSPatchObjf) );
    bsp->udeg = udeg;  bsp->lknu = lknu;
    bsp->vdeg = vdeg;  bsp->lknv = lknv;
    bsp->dim = dim;    bsp->stride = stride;
    ExitIfGLError ( "EnterBSplinePatch 0" );
    glGenBuffers ( 3, bsp->buf );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, bsp->buf[2] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER, bspbsize, NULL, GL_DYNAMIC_DRAW );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[0], sizeof(GLint), &dim );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[1], sizeof(GLint), &udeg );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[2], sizeof(GLint), &lknu );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[3], sizeof(GLint), &vdeg );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[4], sizeof(GLint), &lknv );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[5], sizeof(GLint), &stride );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[6], 3*sizeof(GLfloat), colour );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[7], sizeof(GLint), &one );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[8], sizeof(GLint), &ten );
    ExitIfGLError ( "EnterBSplinePatch 1" );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, bsp->buf[0] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER, (lknu+lknv+2)*sizeof(GLfloat), NULL,
                   GL_DYNAMIC_DRAW );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, 0, (lknu+1)*sizeof(GLfloat), knotsu );
    glBufferSubData ( GL_SHADER_STORAGE_BUFFER, (lknu+1)*sizeof(GLfloat),
                      (lknv+1)*sizeof(GLfloat), knotsv );
    ExitIfGLError ( "EnterBSplinePatch 2" );
    glBindBuffer ( GL_SHADER_STORAGE_BUFFER, bsp->buf[1] );
    glBufferData ( GL_SHADER_STORAGE_BUFFER, stride*(lknu-udeg)*sizeof(GLfloat), cp,
                   GL_DYNAMIC_DRAW );
    ExitIfGLError ( "EnterBSplinePatch 3" );
    glGenVertexArrays ( 1, &bsp->vao );
    glBindVertexArray ( bsp->vao );
    ExitIfGLError ( "EnterBSplinePatch 4" );
    glBindBuffer ( GL_ARRAY_BUFFER, bsp->buf[1] );
    glEnableVertexAttribArray ( 0 );
    glVertexAttribPointer ( 0, dim, GL_FLOAT, GL_FALSE,
                            dim*sizeof(GLfloat), (GLvoid*)0 );
    glBindVertexArray ( 0 );
    ExitIfGLError ( "EnterBSplinePatch" );
  }
  return bsp;
} /*EnterBSplinePatch*/

void SetBSPatchNormals ( BSPatchObjf *bsp, GLint normals )
{
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, bsp->buf[2] );
  glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[7], sizeof(GLint), &normals );
  ExitIfGLError ( "SetBSPatchNormals" );
} /*SetBSPatchNormals*/

void SetBSPatchTessLevel ( BSPatchObjf *bsp, GLint tesslevel )
{
  glBindBuffer ( GL_SHADER_STORAGE_BUFFER, bsp->buf[2] );
  glBufferSubData ( GL_SHADER_STORAGE_BUFFER, bspbofs[8], sizeof(GLint), &tesslevel );
  ExitIfGLError ( "SetBSPatchTessLevel" );
} /*SetBSPatchTessLevel*/

void DrawBSplinePatch ( BSPatchObjf *bsp )
{
  if ( bsp ) {
    glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, bsknbbp, bsp->buf[0] );
    glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, bscpbbp, bsp->buf[1] );
    glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, bspbbp, bsp->buf[2] );
    glBindVertexArray ( empty_vao );
    glPatchParameteri ( GL_PATCH_VERTICES, 1 );
    glDrawArrays ( GL_PATCHES, 0, 1 );
    ExitIfGLError ( "DrawBSplinePatch" );
  }
} /*DrawBSplinePatch*/

void DrawBSplineCPoints ( BSPatchObjf *bsp )
{
  if ( bsp ) {
    glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, bsknbbp, bsp->buf[0] );
    glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, bscpbbp, bsp->buf[1] );
    glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, bspbbp, bsp->buf[2] );
    glBindVertexArray ( bsp->vao );
    glPointSize ( 5.0 );
    glDrawArrays ( GL_POINTS, 0, (bsp->lknu-bsp->udeg)*(bsp->lknv-bsp->vdeg) );
    glBindVertexArray ( 0 );
    ExitIfGLError ( "DrawBSplineCPoints" );
  }
} /*DrawBSplineCPoints*/

void DeleteBSplinePatch ( BSPatchObjf *bsp )
{
  if ( bsp ) {
    glDeleteBuffers ( 3, bsp->buf );
    glDeleteVertexArrays ( 1, &bsp->vao );
    free ( bsp );
  }
} /*DeleteBSplinePatch*/

