#version 450 core

/*#define CALC_POSITION*/

#define MAX_TEXTURES   4
#define MAX_MATERIALS 20
#define MAX_NLIGHTS    8

layout(local_size_x=1) in;

layout(rgba32f,binding=0) uniform image2D  pos;
layout(rgba32f,binding=1) uniform image2D  normal;
layout(rgba32f,binding=2) uniform image2D  tnormal;
layout(rg32f,binding=3)   uniform image2D  txtcoord;
layout(r8i,binding=4)     uniform iimage2D matnum;
layout(rgba32f,binding=6) uniform image2D  glow0;

layout(rgba32f,binding=5) uniform image2D  Out;

layout(binding=0) uniform sampler2D tex[MAX_TEXTURES];

layout(binding=MAX_TEXTURES) uniform sampler2D shtex[MAX_NLIGHTS];
layout(binding=MAX_TEXTURES) uniform samplerCube cshtex[MAX_NLIGHTS];

struct {
     vec3 Position, Normal, TNormal;
     vec2 TxtCoord;
  } In;

uniform TransBlock {
    mat4 mm, mmti, vm, pm, vpm, vpmi;
    vec4 eyepos;
    vec4 viewport;
  } trb;
              
struct LSPar {
    vec3 ambient;
    vec3 direct;
    vec4 position;
    vec3 attenuation;
    mat4 shadow_vpm;
  };

uniform LSBlock {
    uint  nls;              /* liczba zrodel swiatla */
    uint  mask;             /* maska wlaczonych zrodel */
    uint  shmask, cshmask;
    float max_depth;
    int   lightn;
    LSPar ls[MAX_NLIGHTS];  /* poszczegolne zrodla swiatla */
  } light;

struct Material {
    vec4  emission0, emission1;
    vec4  diffref;
    vec4  specref;
    float shininess, wa, we;
    int   txtnum;
  };

uniform MatBlock {
    int mtn;
    Material mat[MAX_MATERIALS];
  } mat;

uniform float glowfct = 2.0;
uniform float emfct = 0.1;
uniform bool  obscgl = false;

Material m;

float CubeDepth ( vec3 v )
{
  v = abs ( v );
  if ( v.x >= v.y ) {
    if ( v.x >= v.z )    return v.x;
    else                 return v.z;
  }
  else if ( v.y >= v.z ) return v.y;
  else                   return v.z;
} /*CubeDepth*/

#ifdef CALC_POSITION
#define OFFSET 0.00001
#else
#define OFFSET 0.0
#endif
float IsEnlightedCube ( uint l )
{
  int   k;
  vec3  lmp;
  float depth;

  depth = CubeDepth ( lmp = In.Position-light.ls[l].position.xyz ) /
          light.max_depth;
  return float ( depth <= texture ( cshtex[l], lmp ).r+OFFSET );
} /*IsEnlightedCube*/

float IsEnlighted ( uint l )
{
  vec4 pos;

  pos = light.ls[l].shadow_vpm * vec4(In.Position,1.0);
  pos.xyz /= pos.w;
  return float ( pos.z <= texture ( shtex[l], pos.xy ).r );
} /*IsEnlighted*/

vec3 posDifference ( vec4 p, vec3 pos, out float dist )
{
  vec3 v;

  if ( p.w != 0.0 ) {
    v = p.xyz/p.w-pos.xyz;
    dist = sqrt ( dot ( v, v ) );
  }
  else
    v = p.xyz;
  return normalize ( v );
} /*posDifference*/

float attFactor ( vec3 att, float dist )
{
  return 1.0/(((att.z*dist)+att.y)*dist+att.x);
} /*attFactor*/

vec3 MatEmission ( float cosnv )
{
  if ( cosnv > 0.0 )
    return emfct * m.emission0.rgb +
           m.emission1.rgb * pow ( cosnv, m.emission1.a );
  else
    return emfct * m.emission0.rgb;
} /*MatEmission*/

vec3 LambertLighting ( void )
{
  vec3  normal, tnormal, lv, vv, Colour;
  float s, d, e, dist;
  uint  i, mask;

  normal = In.Normal;
  tnormal = In.TNormal;
  vv = posDifference ( trb.eyepos, In.Position, dist );
  e = dot ( vv, tnormal );

/*m.diffref.xyz = e > 0.0 ? vec3 (0.5, 0.5, 1.0) : vec3 (0.5, 1.0, 0.5);*/

  Colour = MatEmission ( dot ( normal, vv ) ) /*vec3(0.0)*/;
  for ( i = 0, mask = 0x00000001;  i < light.nls;  i++, mask <<= 1 )
    if ( (light.mask & mask) != 0 ) {
      Colour += light.ls[i].ambient * m.diffref.xyz;
      if ( (light.shmask & mask) != 0 )
        s = IsEnlighted ( i );
      else if ( (light.cshmask & mask) != 0 )
        s = IsEnlightedCube ( i );
      else
        s = 1.0;
      if ( s == 0.0 )
        continue;
      lv = posDifference ( light.ls[i].position, In.Position, dist );
      d = dot ( lv, normal );
      if ( e > 0.0 ) {
        if ( d > 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            d *= attFactor ( light.ls[i].attenuation, dist );
          Colour += (s*d * light.ls[i].direct) * m.diffref.xyz;
        }
      }
      else {
        if ( d < 0.0 ) {
          if ( light.ls[i].position.w != 0.0 )
            d *= attFactor ( light.ls[i].attenuation, dist );
          Colour -= (s*d * light.ls[i].direct) * m.diffref.xyz;
        }
      }
    }
  return clamp ( Colour, 0.0, 1.0 );
} /*LambertLighting*/

#define AGamma(colour) pow ( colour, vec3(256.0/563.0) )

#ifdef CALC_POSITION
vec3 CalcPosition ( vec3 p )
{
  vec4 q;

  q = trb.vpmi *
        vec4 ( 2.0*(p.x-trb.viewport.x)/trb.viewport.z-1.0,
               2.0*(p.y-trb.viewport.y)/trb.viewport.w-1.0,
               2.0*p.z-1.0, 1.0 );
  return q.xyz/q.w;
} /*CalcPosition*/
#endif

void main ( void )
{
  ivec2 xy;
  vec3  Colour;
  vec4  aux0, aux1;
  int   mtn;

  xy = ivec2 ( int(gl_GlobalInvocationID.x), int(gl_GlobalInvocationID.y) );
  aux0 = imageLoad ( glow0, xy );
  Colour = glowfct * aux0.rgb;
  if ( (mtn = imageLoad ( matnum, xy ).x) >= 0 ) {
    aux1 = imageLoad ( normal, xy );
    if ( obscgl && aux1.w < aux0.w )
      Colour = vec3(0.0);
    In.Normal = normalize ( aux1.xyz );
#ifdef CALC_POSITION
    In.Position =
        CalcPosition ( vec3 (float(xy.x)+0.5,float(xy.y)+0.5,aux1.w) );
#else
    In.Position = imageLoad ( pos, xy ).xyz;
#endif
    In.TNormal = imageLoad ( tnormal, xy ).xyz;
    m = mat.mat[mtn];
    if ( m.txtnum >= 0 )
      m.diffref = texture ( tex[m.txtnum], imageLoad ( txtcoord, xy ).xy );
    Colour += LambertLighting ();
  }
  imageStore ( Out, xy, vec4 (AGamma ( Colour ), 1.0) );
} /*main*/

