
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "../utilities/openglheader.h"

/*#define DEBUG_BUFFERS_ALLOCATION*/
#include "../utilities/utilities.h"
#include "deferred.h"
#include "deferredproc.h"
#include "trans.h"
#include "lights.h"
#include "sobject.h"
#include "deferredstruct.h"
#include "room.h"
#include "table.h"
#include "chair.h"
#include "bookcase.h"
#include "galleon.h"
#include "candle.h"

#define MIN_ZOOM 0.1
#define MAX_ZOOM 5.0
#define ZOOM_FCT 1.05

static const float zenith[3] = {0.0,0.0,1.0};

AppData appdata;


static void SetupModelMatrix ( AppData *ad )
{
  M4x4Identf ( ad->trans.mm );
  LoadMMatrix ( &ad->trans, NULL );
} /*SetupModelMatrix*/

static void LoadViewMatrix ( AppData *ad )   
{
/*
printf ( "pos = (%f,%f,%f)\n",
 ad->trans.eyepos[0], ad->trans.eyepos[1], ad->trans.eyepos[2] );
printf ( "angk = %f, rotv = (%f,%f,%f)\n",
 ad->camera.viewer_rangle, ad->camera.viewer_rvec[0],
 ad->camera.viewer_rvec[2], ad->camera.viewer_rvec[2] );
*/
  M4x4InvTranslatefv ( ad->trans.vm, ad->trans.eyepos );
  M4x4RotateVMfv ( ad->trans.vm, ad->camera.viewer_rvec,
                   -ad->camera.viewer_rangle );
  LoadVPMatrix ( &ad->trans );
} /*LoadViewMatrix*/

static void Verticalise ( float vk[3], double *angk )
{
  float  R[16], s, c, vr[3];
  double theta, ang;

  M4x4RotateVfv ( R, vk, *angk );
  s = -V3DotProductf ( &R[0], zenith );
  c = V3DotProductf ( &R[4], zenith );
  theta = atan2 ( s, c );
  V3CompRotationsf ( vr, &ang, &R[8], theta, vk, *angk );
  memcpy ( vk, vr, 3*sizeof(float) );
  *angk = ang;
} /*Verticalise*/

static void InitViewMatrix ( AppData *ad )
{
  const float viewer_pos0[4] = {0.0,0.0,0.0,1.0};
  const float viewer_rvec0[3] = {0.8,0.6,0.0};

  memcpy ( ad->trans.eyepos, viewer_pos0, 4*sizeof(GLfloat) );
  memcpy ( ad->camera.viewer_rvec, viewer_rvec0, 3*sizeof(float) );
  ad->camera.viewer_rangle = 0.5*PI;
  ad->vertical = true;
  Verticalise ( ad->camera.viewer_rvec, &ad->camera.viewer_rangle );
  LoadViewMatrix ( ad );
} /*InitViewMatrix*/

static void _RotateViewer ( AppData *ad,
                            double delta_xi, double delta_eta, char origin )
{
  float  vi[3], lgt, vk[3], vpos[4], z;
  double angi, angk;
  Camera *cam;

  if ( delta_xi == 0 && delta_eta == 0 )
    return;  /* natychmiast uciekamy - nie chcemy dzielic przez 0 */
  cam = &ad->camera;
  vi[0] = (float)delta_eta*(cam->right-cam->left)/(float)cam->win_height;
  vi[1] = (float)delta_xi*(cam->top-cam->bottom)/(float)cam->win_width;
  vi[2] = 0.0;
  lgt = sqrt ( V3DotProductf ( vi, vi ) );
  vi[0] /= lgt;  vi[1] /= lgt;
  z = pow ( ZOOM_FCT, ad->camera.zoomexp );
  angi = 0.0087266463*z;  /* 0.5 stopnia */
  if ( ad->vertical )
    angi *= 0.01 + 0.99*fabs ( ad->trans.vm[1]*zenith[0] +
             ad->trans.vm[5]*zenith[1] + ad->trans.vm[9]*zenith[2] );
  V3CompRotationsf ( vk, &angk, cam->viewer_rvec, cam->viewer_rangle, vi, angi );
  if ( ad->vertical )
    Verticalise ( vk, &angk );
  memcpy ( cam->viewer_rvec, vk, 3*sizeof(float) );
  cam->viewer_rangle = angk;
  if ( origin ) {
    M4x4MultMTV3f ( vk, ad->trans.vm, vi );
    M4x4RotateVfv ( ad->trans.vm, vk, angi );
    M4x4MultMVf ( vpos, ad->trans.vm, ad->trans.eyepos );
    memcpy ( ad->trans.eyepos, vpos, 4*sizeof(float) );
  }
  LoadViewMatrix ( ad );
} /*_RotateViewer*/

void RotateViewer ( double delta_xi, double delta_eta, char origin )
{
  _RotateViewer ( &appdata, delta_xi, delta_eta, origin );
} /*RotateViewer*/

static char _MoveViewer ( AppData *ad, char dir )
{
  static float
    stepb[3] = {0.0,0.0,-0.01}, stepf[3] = {0.0,0.0,0.01},
    stepd[3] = {0.0,-0.01,0.0}, stepu[3] = {0.0,0.01,0.0},
    stepl[3] = {-0.01,0.0,0.0}, stepr[3] = {0.01,0.0,0.0};
  float *w, v[3];
  int i;

  switch ( dir ) {
case MOVE_UP:      w = stepu;  break;
case MOVE_DOWN:    w = stepd;  break;
case MOVE_LEFT:    w = stepl;  break;
case MOVE_RIGHT:   w = stepr;  break;
case MOVE_FORWARD: w = stepf;  break;
case MOVE_BACK:    w = stepb;  break;
default: return false;
  }
  M4x4MultMTV3f ( v, ad->trans.vm, w );
  for ( i = 0; i < 3; i++ )
    ad->trans.eyepos[i] += v[i];
  LoadViewMatrix ( ad );
  return true;
} /*_MoveViewer*/

char MoveViewer ( char dir )
{
  return _MoveViewer ( &appdata, dir );
} /*MoveViewer*/
                  
void ToggleLampCandle ( AppData *ad )
{
  ad->mainlamp_on = !ad->mainlamp_on;
  SetMainLampOnOff ( &ad->lamp, &ad->trans, &ad->mat,
                     &ad->light, 1, ad->mainlamp_on );
  SetCandleOnOff ( &ad->candle, &ad->trans, &ad->mat,
                   &ad->light, 1, !ad->mainlamp_on );
  ad->shadows_ok = false;
} /*ToggleLampCandle*/

static void InitLights ( AppData *ad )
{
  GLfloat amb0[4] = { 0.02, 0.02, 0.02, 0.0 };
  GLfloat dir0[4] = { 0.15, 0.15, 0.15, 0.0 };
  GLfloat pos0[4] = { 0.3, 1.0, 0.55, 0.0 };
  GLfloat atn0[3] = { 1.0, 0.0, 0.0 };

  GLfloat amb2[4] = { 0.024, 0.018, 0.014, 0.0 };
  GLfloat dir2[4] = { 0.36, 0.30, 0.24, 0.0 };
  GLfloat pos2[4] = { -2.95, -2.8, 1.0, 1.0 };
  GLfloat atn2[3] = { 0.3, 0.1, 0.4 };

  GLfloat csc[3] = { 0.0, 0.0, 0.0 };
#define RS 6.5
  LightBl *light;

  light = &ad->light;

  ConstructShadowTxtFBO ( light, 0, false );
  SetLightAmbient ( light, 0, amb0 );
  SetLightDiffuse ( light, 0, dir0 );
  SetLightPosition ( light, 0, pos0 );
  SetLightAttenuation ( light, 0, atn0 );
  SetLightOnOff ( light, 0, true );
  SetupShadowTxtTransformations ( light, 0, csc, RS );
  UpdateShadowMatrix ( light, 0 );

  ConstructShadowTxtFBO ( light, 1, true );
  ad->mainlamp_on = false;
  ToggleLampCandle ( ad );
  SetLightOnOff ( light, 1, true );

  ConstructShadowTxtFBO ( light, 2, true );
  SetLightAmbient ( light, 2, amb2 );
  SetLightDiffuse ( light, 2, dir2 );
  SetLightPosition ( light, 2, pos2 );
  SetLightAttenuation ( light, 2, atn2 );
  SetLightOnOff ( light, 2, true );
  SetupShadowTxtTransformations ( light, 2, csc, RS );

  SetLightMaxDepth ( light, RS );
} /*InitLights*/

static void _ResizeMyWorld ( AppData *ad, int width, int height )
{
  float  lr, z;
  Camera *cam;

  cam = &ad->camera;
  LoadViewport ( &ad->trans, 0, 0,
                 cam->win_width = width, cam->win_height = height );
  z = cam->zoom * 0.05/*0.05533*/;
  lr = z*(float)width/(float)height;  /* przyjmujemy aspekt rowny 1 */
  M4x4Frustumf ( ad->trans.pm, NULL, cam->left = -lr, cam->right = lr,
      cam->bottom = -z, cam->top = z, cam->near = 0.25, cam->far = 16.0 );
  LoadVPMatrix ( &ad->trans );
  ResizeDFBO ( &ad->dfb, width, height );
  ExitIfGLError ( "_ResizeMyWorld" );
} /*_ResizeMyWorld*/

void PrintGLInfo ( void )
{
/*
  GLint n;

  glGetIntegerv ( GL_MAX_COLOR_ATTACHMENTS, &n );
  printf ( "max colour attachments = %d\n", n );
*/
} /*PrintGLInfo*/

void InitMyWorld ( int argc, char *argv[], int width, int height )
{
  memset ( &appdata, 0, sizeof(AppData) );
  PrintGLVersion ();
  PrintGLInfo ();
  LoadMyShaders ( &appdata );
  SetupDFBO ( &appdata.dfb, width, height );
  appdata.trans.trbuf = NewUniformTransBlock ();
  appdata.light.lsbuf = NewUniformLightBlock ();
  appdata.mat.matbuf = NewUniformMatBlock ();
  SetupModelMatrix ( &appdata );
  InitViewMatrix ( &appdata );
  EnterRoomWalls ( &appdata.room, &appdata.mat );
  EnterMainLamp ( &appdata.lamp, &appdata.mat );
  EnterDoor ( &appdata.door, &appdata.mat );
  SetDoorAperture ( &appdata.door, 0.3*PI );
  EnterCarpet ( &appdata.carpet, &appdata.mat );
  EnterChair ( &appdata.chair[0], &appdata.mat );
  appdata.chair[1] = appdata.chair[2] = appdata.chair[0];
  SetChairPosition ( &appdata.chair[0], -1.7, 0.23, -0.15*PI );
  SetChairPosition ( &appdata.chair[1], 0.3, -1.2, 0.55*PI );
  SetChairPosition ( &appdata.chair[2], 0.1, 0.95, -0.6*PI );
  EnterTable ( &appdata.table, &appdata.mat );
  EnterBookcase ( &appdata.bookcase, &appdata.mat );
  SetBookcasePosition ( &appdata.bookcase, 2.98, 1.0, PI );
  EnterGalleon ( &appdata.galleon, &appdata.mat );
  SetGalleonPosition ( &appdata.galleon, -0.4, 0.0,
                       FLOOR_Z+TABLE_HEIGHT+0.28 );
  EnterCandle ( &appdata.candle, &appdata.mat );
  SetCandlePosition ( &appdata.candle, 0.3, 0.25, FLOOR_Z+TABLE_HEIGHT+0.001 );
  InitLights ( &appdata );
  appdata.camera.zoom = 1.0;
  appdata.camera.zoomexp = 0;
  appdata.glowfct = 1.0;
  appdata.ngf = 0;
  appdata.ssao_radius = 0.15;
  _ResizeMyWorld ( &appdata, width, height );
  ExitIfGLError ( "InitMyWorld" );
} /*InitMyObjest*/

void ResizeMyWorld ( int width, int height )
{
  _ResizeMyWorld ( &appdata, width, height );
} /*ResizeMyWorld*/

void RedrawMyWorld ( void )
{
  _RedrawMyWorld ( &appdata );
} /*Redraw*/

static char title0[] = "Cienie w scenie";
static char title1[] = "deferred";
static char title2[] = "glow";
static char title3[] = "SSAO";

char ProcessCharCommand ( char charcode )
{
  switch ( toupper ( charcode ) ) {
case 'L':
    ToggleLampCandle ( &appdata );
    return true;
case '0':
    if ( appdata.render_option != ROPT_DIRECT ) {
      appdata.render_option = ROPT_DIRECT;
      ProcessWorldRequest ( WMSG_SET_TITLE, title0, NULL );
      return true;
    }
    else
      return false;
case '1':
    if ( appdata.render_option != ROPT_DEFERRED1 ) {
      appdata.render_option = ROPT_DEFERRED1;
      ProcessWorldRequest ( WMSG_SET_TITLE, title1, NULL );
      return true;
    }
    else
      return false;
case '2':
    if ( appdata.render_option != ROPT_DEFERRED2 ) {
      appdata.render_option = ROPT_DEFERRED2;
      ProcessWorldRequest ( WMSG_SET_TITLE, title2, NULL );
      return true;
    }
    else
      return false;
case '3':
    if ( appdata.render_option != ROPT_DEFERRED3 ) {
      appdata.render_option = ROPT_DEFERRED3;
      ProcessWorldRequest ( WMSG_SET_TITLE, title3, NULL );
      return true;
    }
    else
      return false;
case 'R':
    InitViewMatrix ( &appdata );
    return true;
case 'V':
    if ( (appdata.vertical = !appdata.vertical) ) {
      Verticalise ( appdata.camera.viewer_rvec, &appdata.camera.viewer_rangle );
      LoadViewMatrix ( &appdata );
      return true;
    }
    else
      return false;
case '+':
    switch ( appdata.render_option ) {
  case ROPT_DEFERRED2:
      if ( appdata.ngf < 12 ) {
        appdata.ngf ++;
        return true;
      }
      else
        return false;
  case ROPT_DEFERRED3:
      if ( appdata.ssao_radius < 1.0 ) {
        appdata.ssao_radius += 0.01;
        printf ( "R = %4.2f\n", appdata.ssao_radius );
        return true;
      }
      else
        return false;
  default:
      return false;
    }
case '-':
    switch ( appdata.render_option ) {
  case ROPT_DEFERRED2:
      if ( appdata.ngf >= 0 ) {
        appdata.ngf --;
        return true;
      }
      else
        return false;
  case ROPT_DEFERRED3:
      if ( appdata.ssao_radius >= 0.02 ) {
        appdata.ssao_radius -= 0.01;
        printf ( "R = %4.2f\n", appdata.ssao_radius );
        return true;
      }
      else
        return false;
  default:
      return false;
    }
case '<':
    if ( appdata.render_option == ROPT_DEFERRED2 ||
         appdata.render_option == ROPT_DEFERRED3 ) {
      appdata.glowfct *= sqrt ( sqrt ( 0.5 ) );
      return true;
    }
    return false;
case '>':
    if ( appdata.render_option == ROPT_DEFERRED2 ||
         appdata.render_option == ROPT_DEFERRED3 ) {
      appdata.glowfct *= sqrt ( sqrt ( 2.0 ) );
      return true;
    }
    return false;
case 'H':
    appdata.obscgl = !appdata.obscgl;
    return true;
case 'C':
    appdata.debug = !appdata.debug;
    return true;
default:
    return false;
  }
} /*ProcessCharCommand*/

char ChangeZoom ( char inczoom )
{
  int    e;
  double z;

  e = appdata.camera.zoomexp;
  if ( inczoom ) {
    if ( (z = pow ( ZOOM_FCT, (double)(++e) )) > MAX_ZOOM )
      return false;
  }
  else {
    if ( (z = pow ( ZOOM_FCT, (double)(--e) )) < MIN_ZOOM )
      return false;
  }
  appdata.camera.zoomexp = e;  appdata.camera.zoom = z;
  _ResizeMyWorld ( &appdata,
                   appdata.camera.win_width, appdata.camera.win_height );
  return true;
} /*ChangeZoom*/

void DeleteMyWorld ( void )
{
  int i;

  for ( i = 0; i < NPROGRAMS; i++ )
    glDeleteProgram ( appdata.program_id[i] );
  glDeleteVertexArrays ( 1, &appdata.room.vao );
  glDeleteBuffers ( 3, appdata.room.vbo );
  glDeleteVertexArrays ( 1, &appdata.lamp.vao );
  glDeleteBuffers ( 3, appdata.lamp.vbo );
  glDeleteVertexArrays ( 1, &appdata.door.vao );
  glDeleteBuffers ( 2, appdata.door.vbo );
  glDeleteVertexArrays ( 1, &appdata.carpet.vao );
  glDeleteBuffers ( 2, appdata.carpet.vbo );
  glDeleteVertexArrays ( 1, &appdata.table.vao );
  glDeleteBuffers ( 2, appdata.table.vbo );
  glDeleteVertexArrays ( 1, &appdata.chair[0].vao );
  glDeleteBuffers ( 2, appdata.chair[0].vbo );
  glDeleteBuffers ( 1, &appdata.trans.trbuf );
  glDeleteBuffers ( 1, &appdata.light.lsbuf );
  glDeleteVertexArrays ( 1, &appdata.bookcase.vao );
  glDeleteBuffers ( 2, appdata.bookcase.vbo );
  glDeleteVertexArrays ( 1, &appdata.galleon.vao );
  glDeleteBuffers ( 2, appdata.galleon.vbo );
  glDeleteVertexArrays ( 1, &appdata.candle.vao );
  glDeleteBuffers ( 2, appdata.candle.vbo );
  DestroyShadowFBO ( &appdata.light );
  DeleteMaterials ( &appdata.mat );
  DeleteDFBO ( &appdata.dfb );
#ifdef DEBUG_BUFFERS_ALLOCATION
  DumpBufferIdentifiers ();
#endif
  ExitIfGLError ( "DeleteMyWorld" );
} /*DeleteMyWorld*/
