#include <stdlib.h>
#include <stdio.h>
#include "openglheader.h" /* najpierw ten */
#include <GL/freeglut.h>  /* potem ten */

#include "utilities.h"
#include "gamma.h"

GLuint shader_id[2]; 
GLuint program_id;
GLuint trbi, trbuf;
GLint  trbsize, trbofs[3];
GLuint uvloc[6];

GLuint vao;

void LoadMyShaders ( void )
{
  static const char *filename[] =
    { "gamma.vert.glsl", "gamma.frag.glsl" };
  static const GLchar *UTBNames[] =
    { "TransBlock", "TransBlock.mm", "TransBlock.vm", "TransBlock.pm" };
  GLuint ind[3];

  shader_id[0] = CompileShaderFiles ( GL_VERTEX_SHADER, 1, &filename[0] );
  shader_id[1] = CompileShaderFiles ( GL_FRAGMENT_SHADER, 1, &filename[1]);
  program_id = LinkShaderProgram ( 2, shader_id, NULL );
  trbi = glGetUniformBlockIndex ( program_id, UTBNames[0] );
  glGetActiveUniformBlockiv ( program_id, trbi,
                              GL_UNIFORM_BLOCK_DATA_SIZE, &trbsize );
  glGetUniformIndices ( program_id, 3, &UTBNames[1], ind );
  glGetActiveUniformsiv ( program_id, 3, ind, GL_UNIFORM_OFFSET, trbofs );
  glUniformBlockBinding ( program_id, trbi, 0 );
  uvloc[0] = glGetUniformLocation ( program_id, "w" );
  uvloc[1] = glGetUniformLocation ( program_id, "h" );
  uvloc[2] = glGetUniformLocation ( program_id, "x" );
  uvloc[3] = glGetUniformLocation ( program_id, "y" );
  uvloc[4] = glGetUniformLocation ( program_id, "c" );
  uvloc[5] = glGetUniformLocation ( program_id, "colour" );

  glGenBuffers ( 1, &trbuf );
  glBindBufferBase ( GL_UNIFORM_BUFFER, 0, trbuf );
  glBufferData ( GL_UNIFORM_BUFFER, trbsize, NULL, GL_DYNAMIC_DRAW );
  ExitIfGLError ( "LoadMyShaders" );
} /*LoadMyShaders*/

void InitMyObject ( void )
{
  ConstructVAO ();
} /*InitMyObject*/

void ConstructVAO ( void )
{
  glGenVertexArrays ( 1, &vao );
  glVertexAttrib3f ( 0, 0.0, 0.0, 0.0 );
  ExitIfGLError ( "ConstructVAO" );
} /*ConstructVAO*/

void DrawRect ( GLint w, GLint h, GLint x, GLint y, GLint c, GLfloat colour[3] )
{
  glUseProgram ( program_id );
  glUniform1i ( uvloc[0], w );
  glUniform1i ( uvloc[1], h );
  glUniform1i ( uvloc[2], x );
  glUniform1i ( uvloc[3], y );
  glUniform1i ( uvloc[4], c );
  glUniform3fv ( uvloc[5], 1, colour );
  glBindVertexArray ( vao );
  glDrawArrays ( GL_TRIANGLE_FAN, 0, 4 );
} /*DrawRect*/

void DrawIt ( void )
{
#define CC 0.75
  GLfloat colour[21][3] =
    {{0.5,0.0,0.0},{1.0,0.0,0.0},{CC,0.0,0.0},
     {0.5,0.5,0.0},{1.0,1.0,0.0},{CC,CC,0.0},
     {0.0,0.5,0.0},{0.0,1.0,0.0},{0.0,CC,0.0},
     {0.0,0.5,0.5},{0.0,1.0,1.0},{0.0,CC,CC},
     {0.0,0.0,0.5},{0.0,0.0,1.0},{0.0,0.0,CC},
     {0.5,0.0,0.5},{1.0,0.0,1.0},{CC,0.0,CC},
     {0.5,0.5,0.5},{1.0,1.0,1.0},{CC,CC,CC}};
  glUseProgram ( program_id );
  glBindVertexArray ( vao );
  DrawRect ( 1, 1, 0, 0, false, colour[0] );
  DrawRect ( 1, 1, 0, 1, true, colour[1] );
  DrawRect ( 1, 1, 0, 2, false, colour[2] );
  DrawRect ( 1, 1, 1, 0, false, colour[3] );
  DrawRect ( 1, 1, 1, 1, true, colour[4] );
  DrawRect ( 1, 1, 1, 2, false, colour[5] );
  DrawRect ( 1, 1, 2, 0, false, colour[6] );
  DrawRect ( 1, 1, 2, 1, true, colour[7] );
  DrawRect ( 1, 1, 2, 2, false, colour[8] );
  DrawRect ( 1, 1, 3, 0, false, colour[9] );
  DrawRect ( 1, 1, 3, 1, true, colour[10] );
  DrawRect ( 1, 1, 3, 2, false, colour[11] );
  DrawRect ( 1, 1, 4, 0, false, colour[12] );
  DrawRect ( 1, 1, 4, 1, true, colour[13] );
  DrawRect ( 1, 1, 4, 2, false, colour[14] );
  DrawRect ( 1, 1, 5, 0, false, colour[15] );
  DrawRect ( 1, 1, 5, 1, true, colour[16] );
  DrawRect ( 1, 1, 5, 2, false, colour[17] );
  DrawRect ( 1, 1, 6, 0, false, colour[18] );
  DrawRect ( 1, 1, 6, 1, true, colour[19] );
  DrawRect ( 1, 1, 6, 2, false, colour[20] );

  glBindVertexArray ( 0 );
} /*DrawIcosahedron*/

void Cleanup ( void )
{
  int i;

  glUseProgram ( 0 );
  for ( i = 0; i < 2; i++ )
    glDeleteShader ( shader_id[i] );
  glDeleteProgram ( program_id );
  glDeleteBuffers ( 1, &trbuf );
  glDeleteVertexArrays ( 1, &vao );
  ExitIfGLError ( "Cleanup" );
  glutDestroyWindow ( WindowHandle );
} /*Cleanup*/

