
#include <stdlib.h>
#include <stdio.h>
#include "openglheader.h"

#include "utilities.h"
#include "GPUsort.h"

GLuint LoadGPUNetSortShaders ( const char *cs_filename, GLuint *uvloc )
{
  static const char *filename = "sortnet.comp.glsl";
  static const GLchar *uname[] = { "n", "n0", "reverse", "h" };
  GLuint shader_id[2], program_id;
  int i;

  shader_id[0] = CompileShaderFiles ( GL_COMPUTE_SHADER, 1, &filename );
  shader_id[1] = CompileShaderFiles ( GL_COMPUTE_SHADER, 1, &cs_filename );
  program_id = LinkShaderProgram ( 2, shader_id, "GPUNetSort" );
  for ( i = 0; i < 4; i++ )
    uvloc[i] = glGetUniformLocation ( program_id, uname[i] );
  for ( i = 0; i < 2; i++ )
    glDeleteShader ( shader_id[i] );
  ExitIfGLError ( "LoadGPUNetSortShaders" );
  return program_id;
} /*LoadGPUNetSortShaders*/

void GPUNetSort ( unsigned int n, unsigned int n0, GLuint dbuf,
                  GLuint program_id, GLuint *uvloc )
{
  unsigned int steps, nn, h, h2, gsize, i;

  if ( n < 2 )
    return;
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 0, dbuf );
  glUseProgram ( program_id );
  for ( nn = n-1, steps = 0;  nn;  nn >>= 1, steps ++ )
    ;
  nn = 1 << steps;  gsize = nn/2;
  glUniform1ui ( uvloc[0], n );
  glUniform1ui ( uvloc[1], n0 );
  for ( i = 0, h2 = 1, h = 2;  i < steps;  i++, h2 = h, h += h ) {
    glUniform1ui ( uvloc[3], h );
    glUniform1ui ( uvloc[2], true );
    glDispatchCompute ( gsize, 1, 1 );
    glMemoryBarrier ( GL_SHADER_STORAGE_BARRIER_BIT );
    glUniform1ui ( uvloc[2], false );
    for ( ;  h2 > 1;  h2 >>= 1 ) {
      glUniform1ui ( uvloc[3], h2 );
      glDispatchCompute ( gsize, 1, 1 );
      glMemoryBarrier ( GL_SHADER_STORAGE_BARRIER_BIT );
    }
  }
  ExitIfGLError ( "GPUNetSort" );
} /*GPUNetSort*/

