
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "openglheader.h"
#include <GLFW/glfw3.h>

#include "utilities.h"
#include "GPUsort.h"

GLFWwindow *mywindow;
int win_width, win_height;

char redraw = 0;

void myGLFWErrorHandler ( int error, const char *description )
{
  fprintf ( stderr, "GLFW error: %s\n", description );
  exit ( 1 );
} /*myGLFWErrorHandler*/

/* ////////////////////////////////////////////////////////////////////////// */
void CompSwap ( unsigned int n, void *data, unsigned int i, unsigned int j )
{
  int *d, x;

  if ( j < n ) {
    d = (int*)data;
    if ( d[i] > d[j] )
      { x = d[i];  d[i] = d[j];  d[j] = x; }
  }
} /*CompSwap*/

void NetSort ( unsigned int n, void *data )
{
  unsigned int steps, nn, h, hh, hhh, i, j, k, kk, l;

  if ( n < 2 )
    return;
  for ( nn = n-1, steps = 0;  nn;  nn >>= 1, steps ++ )
    ;
  nn = 1 << steps;
  k = nn/2;
  for ( j = 0; j < k; j++ )  /* in parallel */
    CompSwap ( n, data, j+j, j+j+1 );
  for ( i = 1, hh = 2, h = 4, k = nn >> 2;
        i < steps;
        i++, hh = h, h <<= 1, k >>= 1 ) {
    for ( j = 0; j < k; j++ ) { /* in parallel */
      for ( l = 0; l < hh; l++ )
        CompSwap ( n, data, j*h+l, j*h+h-1-l );
    }
    for ( hh = h >> 1, hhh = hh >> 1, kk = k << 1;
          hh > 1;
          hh = hhh, kk <<= 1, hhh >>= 1 ) {
      for ( j = 0; j < kk; j++ ) {  /* in parallel */
        for ( l = 0; l < hhh; l++ )
          CompSwap ( n, data, j*hh+l, j*hh+l+hhh );
      }
    }
  }
} /*NetSort*/

/* ////////////////////////////////////////////////////////////////////////// */
#define N 62
int data[N] = {0,33,101,37,5,99,20,9,19,87,36,6,88,58,10,2,54,76,11,57,105,15,
               102,52,121,16,63,73,19,67,51,65,80,110,25,42,98,61,75,28,163,
               30,3,35,7,39,20,39,16,56,8,50,2,4,51,17,55,12,46,13,59,1};
int sdata[N];

void ShowData ( unsigned int n, void *data, unsigned int h )
{
  int *d;
  int i;

  d = (int*)data;
  for ( i = 0; i < n; i++ ) {
    printf ( "%4d,", d[i] );
    if ( !((i+1) % h) )
      printf ( "\n" );
  }
  printf ( "\n" );
} /*ShowData*/

void InitMyObject ( void )
{
  GLuint program_id;
  GLuint uvloc[4];
  GLuint  dbuf;
  
  program_id = LoadGPUNetSortShaders ( "sortnetcs.comp.glsl", uvloc );
  glGenBuffers ( 1, &dbuf );
  glBindBufferBase ( GL_SHADER_STORAGE_BUFFER, 0, dbuf );
  glBufferData ( GL_SHADER_STORAGE_BUFFER, N*sizeof(GLint), data, GL_DYNAMIC_READ );
  GPUNetSort ( N, 0, dbuf, program_id, uvloc );
  glGetBufferSubData ( GL_SHADER_STORAGE_BUFFER, 0, N*sizeof(GLint), sdata );
  ShowData ( N, sdata, 16 );
  glDeleteBuffers ( 1, &dbuf );
  glDeleteProgram ( program_id );
  ExitIfGLError ( "InitMyObject" );
} /*InitMyObject*/

void ReshapeFunc ( GLFWwindow *win, int width, int height )
{
} /*ReshapeFunc*/

void Redraw ( GLFWwindow *win )
{
  glClearColor ( 1.0, 1.0, 1.0, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glUseProgram ( 0 );
  glFlush ();
  ExitIfGLError ( "Redraw" );
  redraw = false;
  glfwSwapBuffers ( win );
} /*Redraw*/

void DisplayFunc ( GLFWwindow *win )
{
  redraw = true; 
} /*DisplayFunc*/

void KeyFunc ( GLFWwindow *win, int key, int scancode, int action, int mode )
{
  if ( action == GLFW_PRESS || action == GLFW_REPEAT ) {
    switch ( key ) {   
  case GLFW_KEY_ESCAPE:
      glfwSetWindowShouldClose ( mywindow, 1 );
      break;
  default:  
      break;
    }
  }
} /*KeyFunc*/

void CharFunc ( GLFWwindow *win, unsigned int charcode )
{
} /*CharFunc*/

void MouseFunc ( GLFWwindow *win, int button, int action, int mods )
{
} /*MouseFunc*/

void MotionFunc ( GLFWwindow *win, double x, double y )
{
} /*MotionFunc*/

void ScrollFunc ( GLFWwindow *win, double x, double y )
{
} /*ScrollFunc*/

void IdleFunc ( void )
{
} /*IdleFunc*/

void Initialise ( int argc, char **argv )
{
  glfwSetErrorCallback ( myGLFWErrorHandler );
  if ( !glfwInit () )
    ExitOnError ( "Initialise 0" );
  if ( !(mywindow = glfwCreateWindow ( 480, 360,
                                       "Sortnet", NULL, NULL )) ) {
    glfwTerminate ();
    ExitOnError ( "Initialise 1" );
  }
  glfwMakeContextCurrent ( mywindow );
  GetGLProcAddresses ( 4, 3 );
  PrintGLVersion ();
  glfwSetFramebufferSizeCallback ( mywindow, ReshapeFunc );
  glfwSetWindowRefreshCallback ( mywindow, DisplayFunc );
  glfwSetKeyCallback ( mywindow, KeyFunc );  
  glfwSetCharCallback ( mywindow, CharFunc );
  glfwSetMouseButtonCallback ( mywindow, MouseFunc );
  glfwSetCursorPosCallback ( mywindow, MotionFunc );
  glfwSetScrollCallback ( mywindow, ScrollFunc );
  InitMyObject ();
  ReshapeFunc ( mywindow, 480, 360 );
  redraw = true;
  ExitIfGLError ( "Initialise" );
} /*Initialise*/

void MessageLoop ( void )
{
  do {
    glfwWaitEvents ();
    if ( redraw )
      Redraw ( mywindow );
  } while ( !glfwWindowShouldClose ( mywindow ) );
} /*MessageLoop*/

void Cleanup ( void )
{
} /*Cleanup*/

int main ( int argc, char **argv )
{
  Initialise ( argc, argv );
  MessageLoop ();
  Cleanup ();
  exit ( 0 );
} /*main*/

