#version 440 core

#define INF0 65533
#define INF1 65534
#define INF2 65535
#define R        5.0
#define RR      25.0

uniform TransBlock {
    dvec4 rxy;
    dvec2 wh;
    int   width, mag, maxit, currit, diter;
  } tr;

layout(r32f,binding=1) uniform image2D img;
layout(std430,binding=0) buffer Cmap { dvec2 cmap[]; };

void main ( void )
{
  dvec2  c, z;
  double xx, x2, y2, r2, rk2, fr;
  int    i;
  ivec2  xy;

  xy = ivec2 ( int(gl_FragCoord.x), int(gl_FragCoord.y) );
  c = dvec2 ( tr.rxy.x + double(gl_FragCoord.x)/tr.wh.x*(tr.rxy.y-tr.rxy.x),
              tr.rxy.z + double(gl_FragCoord.y)/tr.wh.y*(tr.rxy.w-tr.rxy.z) );
  if ( tr.currit == 0 ) {
    xx = c.x+1.0;  y2 = c.y*c.y;
    if ( xx*xx+y2 <= 0.0625 ) {
      imageStore ( img, xy, vec4 ( INF0, 0.0, 0.0, 0.0 ) );
      cmap[xy.y*tr.width + xy.x] = c;
      discard;
    }
    else {
      xx = c.x - 0.25;  x2 = xx*xx;  r2 = x2+y2;
      if ( (r2 + xx)*r2 <= 0.25 * y2 ) {
        imageStore ( img, xy, vec4 ( INF1, 0.0, 0.0, 0.0 ) );
        cmap[xy.y*tr.width + xy.x] = c;
        discard;
      }
    }
    imageStore ( img, xy, vec4 ( INF2, 0.0, 0.0, 0.0 ) );
    z = dvec2 ( 0.0, 0.0 );
    rk2 = 0.0;
  }
  else {
    if ( imageLoad ( img, xy ).r < INF0 )
      discard;
    z = cmap[xy.y*tr.width + xy.x];
    rk2 = z.x*z.x + z.y*z.y;
  }

  for ( i = tr.currit;  i < tr.currit+tr.diter;  i++ ) {
    z = dvec2 ( (z.x + z.y)*(z.x - z.y) + c.x, (z.x + z.x)*z.y + c.y );
    if ( (r2 = z.x*z.x + z.y*z.y) > RR ) {
      fr = (R-sqrt ( rk2 )) / (sqrt ( r2 )-sqrt ( rk2 ));
      imageStore ( img, xy, vec4 ( float(i) + fr, 0.0, 0.0, 0.0 ) );
      cmap[xy.y*tr.width + xy.x] = z;
      discard;
    }
    rk2 = r2;
  }
  cmap[xy.y*tr.width + xy.x] = z;
  discard;
} /*main*/

