#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "math.h"
#include "openglheader.h"

#include "utilities.h"
#include "mygltext.h"
#include "offsfbo.h"
#include "mandelbrot.h"

static OffsFBO *fbo;
int            wdt, hgh, magaa = 1;
static double  width, height;
static double  xc = -0.5, yc = 0.0, rc = 3.0;
static double  dxc, dyc;
static char    mappingchanged = true, finished = false;

static GLuint program_id[2];
static GLuint trbuf, trbpoint;
static GLint  trofs[7];
static GLuint palbuf, palbpoint;
static GLint  palofs[3];
static int    maxitcnt = 0;
static GLint  maxiter = MINMAXIT, curriter = 0, diter = 100;

static myTextObject *mytext;
static myFont *fonts[2], *font;
static GLint  pn_loc;

/* ////////////////////////////////////////////////////////////////////////// */
#define MAXDEG 10
static GLfloat rho = 1.0;
static GLint   BezPaletteDeg = 2;
static GLfloat BezPaletteCP0[MAXDEG+1][4] =
    {{1.0,0.0,1.0,1.0},{0.0,1.0,1.0,1.0},{0.0,0.0,1.0,1.0}};
static GLfloat BPCP[MAXDEG+1][4], BezPaletteCP[MAXDEG+1][4];
static const GLchar *upalname[] =
  { "PaletteBlock", "PaletteBlock.deg", "PaletteBlock.cp", "PaletteBlock.bpcp" };
static GLfloat fg0[4] = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat bk0[4] = { 0.0, 0.0, 0.0, 1.0 };
static GLfloat fg1[4] = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat bk1[4] = { 0.0, 0.0, 0.0, 0.0 };

void LoadMyShaders1 ( void )
{
  static const GLchar *shfname[] = { "mand.vert.glsl",
    "mand1.frag.glsl", "mand2.frag.glsl" };
  static const GLchar *utrname[] =
    { "TransBlock", "TransBlock.rxy", "TransBlock.wh",
      "TransBlock.maxit", "TransBlock.mag", "TransBlock.currit",
      "TransBlock.width", "TransBlock.diter" };
  static const GLchar *uvnames[] = { "pn" };
  GLuint shader_id[3], shid[2], i;
  GLint  size;

  shader_id[0] = CompileShaderFiles ( GL_VERTEX_SHADER, 1, &shfname[0] );
  shader_id[1] = CompileShaderFiles ( GL_FRAGMENT_SHADER, 1, &shfname[1] );
  shader_id[2] = CompileShaderFiles ( GL_FRAGMENT_SHADER, 1, &shfname[2] );
  program_id[0] = LinkShaderProgram ( 2, shader_id, "0" );
  shid[0] = shader_id[0]; shid[1] = shader_id[2];
  program_id[1] = LinkShaderProgram ( 2, shid, "1" );
/*PrintProgramResources ( program_id[0], "0" );
PrintProgramResources ( program_id[1], "1" );*/
  pn_loc = glGetUniformLocation ( program_id[1], uvnames[0] );
  GetAccessToUniformBlock ( program_id[0], 7, utrname, &size, trofs, &trbpoint );
  AttachUniformBlockToBP ( program_id[1], utrname[0], trbpoint );
  trbuf = NewUniformBuffer ( size, trbpoint );
  GetAccessToUniformBlock ( program_id[1], 3, upalname, &size, palofs, &palbpoint );
  palbuf = NewUniformBuffer ( size, palbpoint );
  for ( i = 0; i < 3; i++ )
    glDeleteShader ( shader_id[i] );
  ExitIfGLError ( "LoadMyShaders1" );
} /*LoadMyShaders1*/

void DeleteMyShaders1 ( void )
{
  int i;

  glUseProgram ( 0 );
  for ( i = 0; i < 2; i++ )
    glDeleteProgram ( program_id[i] );
} /*DeleteMyShaders1*/

void LoadMyPalette ( void )
{
  glBindBuffer ( GL_UNIFORM_BUFFER, palbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, palofs[0], sizeof(GLint), &BezPaletteDeg );
  glBufferSubData ( GL_UNIFORM_BUFFER, palofs[1],
                    (BezPaletteDeg+1)*4*sizeof(GLfloat), BezPaletteCP );
  glBufferSubData ( GL_UNIFORM_BUFFER, palofs[2],
                    (BezPaletteDeg+1)*4*sizeof(GLfloat), BPCP );
  ExitIfGLError ( "LoadMyPalette" );
} /*LoadMyPalette*/

void PaletteReset ( void )
{
  BezPaletteDeg = 2;
  rho = 1.0;
  memcpy ( BPCP, BezPaletteCP0, 12*sizeof(GLfloat) );
  memcpy ( BezPaletteCP, BezPaletteCP0, 12*sizeof(GLfloat) );
  LoadMyPalette ();
} /*PaletteReset*/

void InitMyObject1 ( void )
{
  TimerInit ();
  LoadMyShaders1 ();
  LoadTextShaders ();
  PaletteReset ();
  fbo = NewOffsFBO ( 480*MAGAA, 360*MAGAA );
  ConstructEmptyVAO ();
  glMinSampleShading ( 1.0 );
  glDisable ( GL_SAMPLE_SHADING );
  font = fonts[0] = NewFont18x10 ();
  fonts[1] = NewFont12x6 ();
  mytext = NewTextObject ( 128 );
  SetTextForeground ( fg1 );
  SetTextBackground ( bk1 );
} /*InitMyObject1*/

void DeleteMyObject1 ( void )
{
  DeleteMyShaders1 ();
  DeleteOffsFBO ( fbo );
  DeleteFontObject ( fonts[0] );
  DeleteFontObject ( fonts[1] );
  TextCleanup ();
  DeleteEmptyVAO ();
  glDeleteBuffers ( 1, &trbuf );
  glDeleteBuffers ( 1, &palbuf );
  DeleteTextObject ( mytext );
} /*DeleteMyObject1*/

void NotifyRectangle ( void )
{
  char nd;
  char s[128];

  nd = 3 - floor ( log10 ( rc ) );
  sprintf ( s, "xc = %.*f, yc = %.*f, r = %.3g, N = %d",
            nd, xc, nd, yc, rc, maxiter );
  SetTextObjectContents ( mytext, s, 0, 18, font );
} /*NotifyRectangle*/

void FindTheMapping ( int w, int h )
{
  double   d, s, c;
  GLdouble xyc[4];

  if ( !SetOffsFBOSize ( fbo, w*MAGAA, h*MAGAA ) )
    ExitOnError ( "FindTheMapping" );
  width = (double)(wdt = w);  height = (double)(hgh = h);
  d = sqrt ( width*width + height*height );
  s = height/d;  c = width/d;
  dxc = rc*c;     dyc = rc*s;
  xyc[0] = xc - dxc;  xyc[1] = xc + dxc;
  xyc[2] = yc - dyc;  xyc[3] = yc + dyc;
  glBindBuffer ( GL_UNIFORM_BUFFER, trbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, trofs[0], 4*sizeof(GLdouble), xyc );
  NotifyRectangle ();
  mappingchanged = true;
/*printf ( "+" );*/
  ExitIfGLError ( "FindTheMapping" );
} /*FindTheMapping*/

char Zoom ( char in )
{
  double r;

  if ( in ) {
    if ( (r = rc / 1.05) >= MIN_RC )
      rc = r;
    else
      return false;
  }
  else {
    if ( (r = rc * 1.05) <= MAX_RC )
      rc = r;
    else
      return false;
  }
  FindTheMapping ( wdt, hgh );
  return true;
} /*Zoom*/

char Pan ( double dx, double dy )
{
  xc -= (dx/width)*2.0*dxc;  yc += (dy/height)*2.0*dyc;
  FindTheMapping ( wdt, hgh );
  return true;
} /*Pan*/

void ResetTheMapping ( void )
{
  xc = -0.5, yc = 0.0, rc = 3.0;
  FindTheMapping ( wdt, hgh );
} /*ResetTheMapping*/

char ChangeMaxIter ( char inc )
{
  int maxit;

  maxit = maxiter;
  if ( inc && maxiter < MAXMAXIT ) {
    maxitcnt ++;
    goto setmaxiter;
  }
  else if ( !inc && maxitcnt > 0 ) {
    maxitcnt --;
setmaxiter:
    maxiter = (GLint)(MINMAXIT*pow ( DRHO, maxitcnt ));
    if ( maxiter > MAXMAXIT )
      maxiter = MAXMAXIT;
    if ( maxiter != maxit )
      FindTheMapping ( wdt, hgh );
  }
  return maxiter != maxit;
} /*ChangeMaxIter*/

void ChoosePalette ( int np )
{
  glUseProgram ( program_id[1] );
  glUniform1i ( pn_loc, np );
  switch ( np ) {
case 1:  case 2:
    SetTextForeground ( fg1 );
    SetTextBackground ( bk1 );
    break;
default:
    SetTextForeground ( fg0 );
    SetTextBackground ( bk0 );
    break;
  }
} /*ChoosePalette*/

char RedrawMyObject1 ( void )
{
  GLdouble xyc[2];
  GLint    dit;
  double   t0, t1, di;

  if ( mappingchanged ) { finished = false;  curriter = 0; }
  if ( !finished ) {
    magaa = antialias1 ? MAGAA : 1;
    xyc[0] = magaa*width;  xyc[1] = magaa*height;
    glBindBufferBase ( GL_UNIFORM_BUFFER, trbpoint, trbuf );
    glBufferSubData ( GL_UNIFORM_BUFFER, trofs[1], 2*sizeof(GLdouble), xyc );
    glBufferSubData ( GL_UNIFORM_BUFFER, trofs[2], sizeof(GLint), &maxiter );
    glBufferSubData ( GL_UNIFORM_BUFFER, trofs[3], sizeof(GLint), &magaa );
    glBufferSubData ( GL_UNIFORM_BUFFER, trofs[4], sizeof(GLint), &curriter );
    glBufferSubData ( GL_UNIFORM_BUFFER, trofs[5], sizeof(GLint), &fbo->width );
    dit = curriter+diter <= maxiter ? diter : maxiter-curriter;
    glBufferSubData ( GL_UNIFORM_BUFFER, trofs[6], sizeof(GLint), &dit );
    glBindFramebuffer ( GL_FRAMEBUFFER, fbo->fbo );
    glViewport ( 0, 0, magaa*wdt, magaa*hgh );
    glBindImageTexture ( 1, fbo->txt, 0, GL_FALSE, 0, GL_WRITE_ONLY, GL_R32F );
    glUseProgram ( program_id[0] );
    glBindVertexArray ( empty_vao );
    glFinish ();
    t0 = TimerToc ();
    glDrawArrays ( GL_TRIANGLE_FAN, 0, 4 );
    mappingchanged = false;
    glFinish ();
    t1 = TimerToc ();
    di = (double)dit*FTIME/(t1-t0);
    diter = di < MINMAXIT ? MINMAXIT : (di > MAXMAXIT ? MAXMAXIT : (int)di);
/*printf ( "diter = %d, t1-t0 = %9.6f\n", diter, t1-t0 );*/
    glBindFramebuffer ( GL_FRAMEBUFFER, 0 );
    finished = (curriter += dit) >= maxiter;
  }
  glViewport ( 0, 0, wdt, hgh );
  glBindImageTexture ( 1, fbo->txt, 0, GL_FALSE, 0, GL_READ_ONLY, GL_R32F );
  glUseProgram ( program_id[1] );
  glDrawArrays ( GL_TRIANGLE_FAN, 0, 4 );
  glBindVertexArray ( 0 );
  if ( showtext1 )
    DisplayTextObject ( mytext );
  glUseProgram ( 0 );
  glFlush ();
  ExitIfGLError ( "RedrawMyObject1" );
  return finished;
} /*RedrawMyObject1*/

/* ////////////////////////////////////////////////////////////////////////// */
GLuint  program2_id[6];
GLint   triangofs[2];
GLuint  triangbpoint, triangbuf;
GLint   win2w, win2h;
GLfloat win2vpm[16], win2vpminv[16], win2sqvpm[16], win2sqvpminv[16];
int     current;

void LoadMyShaders2 ( void )
{
  static const GLchar *shfname[] = { "triang.vert.glsl", "triang.frag.glsl",
    "bezc.vert.glsl", "bezc.tesc.glsl", "bezc.tese.glsl", "bezc.frag.glsl",
    "bezccp.vert.glsl", "square.vert.glsl", "square.frag.glsl",
    "bpolygon.vert.glsl", "bpoly.tese.glsl" };
  static const GLenum shtype[] =
    { GL_VERTEX_SHADER, GL_FRAGMENT_SHADER,
      GL_VERTEX_SHADER, GL_TESS_CONTROL_SHADER, GL_TESS_EVALUATION_SHADER,
      GL_FRAGMENT_SHADER, GL_VERTEX_SHADER,
      GL_VERTEX_SHADER, GL_FRAGMENT_SHADER, GL_VERTEX_SHADER,
      GL_TESS_EVALUATION_SHADER };
  static const GLchar *utriangname[] =
    { "TriangBlock", "TriangBlock.vpm", "TriangBlock.sqpm" };
  GLuint shader_id[9], shid[4];
  int    i;
  GLint  size;

  for ( i = 0; i < 11; i++ )
    shader_id[i] = CompileShaderFiles ( shtype[i], 1, &shfname[i] );
  program2_id[0] = LinkShaderProgram ( 2, shader_id, "3" );
  program2_id[1] = LinkShaderProgram ( 4, &shader_id[2], "4" );
  shid[0] = shader_id[6];  shid[1] = shader_id[5];
  program2_id[2] = LinkShaderProgram ( 2, shid, "5" );
  program2_id[3] = LinkShaderProgram ( 2, &shader_id[7], "6" );
  shid[0] = shader_id[9];  shid[1] = shader_id[5];
  program2_id[4] = LinkShaderProgram ( 2, shid, "7" );
  shid[0] = shader_id[2];  shid[1] = shader_id[3];  shid[2] = shader_id[10];
  shid[3] = shader_id[5];
  program2_id[5] = LinkShaderProgram ( 4, shid, "8" );
  GetAccessToUniformBlock ( program2_id[0], 2, utriangname,
                            &size, triangofs, &triangbpoint );
  triangbuf = NewUniformBuffer ( size, triangbpoint );
  AttachUniformBlockToBP ( program2_id[1], upalname[0], palbpoint );
  AttachUniformBlockToBP ( program2_id[2], upalname[0], palbpoint );
  AttachUniformBlockToBP ( program2_id[1], utriangname[0], triangbpoint );
  AttachUniformBlockToBP ( program2_id[2], utriangname[0], triangbpoint );
  AttachUniformBlockToBP ( program2_id[3], utriangname[0], triangbpoint );
  AttachUniformBlockToBP ( program2_id[4], upalname[0], palbpoint );
  AttachUniformBlockToBP ( program2_id[4], utriangname[0], triangbpoint );
  AttachUniformBlockToBP ( program2_id[5], upalname[0], palbpoint );
  AttachUniformBlockToBP ( program2_id[5], utriangname[0], triangbpoint );
  for ( i = 0; i < 11; i++ )
    glDeleteShader ( shader_id[i] );
  ExitIfGLError ( "LoadMyShaders2" );
} /*LoadMyShaders2*/

void DeleteMyShaders2 ( void )
{
  glDeleteProgram ( program2_id[0] );
} /*DeleteMyShaders2*/

void InitMyObject2 ( void )
{
  LoadMyShaders2 ();
  PaletteReset ();
} /*InitMyObject2*/

void DeleteMyObject2 ( void )
{
  DeleteMyShaders2 ();
  glDeleteBuffers ( 1, &triangbuf );
} /*DeleteMyObject2*/

void FindMapping2 ( int width, int height )
{
  GLfloat pm[16], vm[16];
  GLfloat l, r, b, t, z;

  win2w = width;  win2h = height;
  l = 0.0;  b = 0.0;
  if ( width/2 < height ) {
    r = 2.0;  t = (float)(2*height)/(float)width;
  }
  else {
    r = (float)width/(float)height; t = 1.0;
  } 
  l = -0.1*r;  r *= 1.1;
  b = -0.1*t;  t *= 1.1;
  M4x4Orthof ( pm, NULL, l, r, b, t, -1.0, 1.0 );
  M4x4Identf ( vm );
  vm[4] = 0.5;  vm[5] = sqrt(3.0)/2.0;
  M4x4Multf ( win2vpm, pm, vm );
  M4x4Invertf ( win2vpminv, win2vpm );
  glBindBufferBase ( GL_UNIFORM_BUFFER, triangbpoint, triangbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, triangofs[0],
                    16*sizeof(GLfloat), win2vpm );
  z = (float)width/(float)height;
  if ( width/2 < height ) {
    l = -11.0/8.0;  r = 9.0/8.0;
    b = -1.0/8.0;   t = (20.0/z - 1.0)/8.0;
  }
  else {
    l = -(1.0+5.0*z)/8.0;  r = (5.0*z-1.0)/8.0;
    b = -1.0/8.0;          t = b+1.0/0.8;
  }
  M4x4Orthof ( win2sqvpm, win2sqvpminv, l, r, b, t, -1.0, 1.0 );
  glBufferSubData ( GL_UNIFORM_BUFFER, triangofs[1],
                    16*sizeof(GLfloat), win2sqvpm );
  ExitIfGLError ( "FindMapping2" );
} /*FindMapping2*/

void RedrawMyObject2 ( void )
{
  glViewport ( 0, 0, win2w, win2h );
  glClearColor ( 0.33, 0.33, 0.33, 1.0 );
  glClear ( GL_COLOR_BUFFER_BIT );
  glUseProgram ( program2_id[0] );
  glVertexAttrib1f ( 0, 0.0 );
  glDrawArrays ( GL_TRIANGLES, 0, 3 );
  glUseProgram ( program2_id[1] );
  glBindBufferBase ( GL_UNIFORM_BUFFER, palbpoint, palbuf );
  glPatchParameteri ( GL_PATCH_VERTICES, 2 );
  glDrawArrays ( GL_PATCHES, 0, 2 );
  glUseProgram ( program2_id[2] );
  glPointSize ( 5.0 );
  glDrawArrays ( GL_POINTS, 0, BezPaletteDeg+1 );
  glDrawArrays ( GL_LINE_STRIP, 0, BezPaletteDeg+1 );
  glUseProgram ( program2_id[3] );
  glDrawArrays ( GL_TRIANGLE_FAN, 0, 4 );
  glUseProgram ( program2_id[4] );
  glDrawArrays ( GL_POINTS, 0, BezPaletteDeg+1 );
  glDrawArrays ( GL_LINE_STRIP, 0, BezPaletteDeg+1 );
  glUseProgram ( program2_id[5] );
  glDrawArrays ( GL_PATCHES, 0, 2 );
  glUseProgram ( 0 );
  glFlush ();
  ExitIfGLError ( "RedrawMyObject2" );
} /*RedrawMyObject2*/

static void SetPaletteRho ( float r )
{
  int   i, j;
  float ri;

  rho = r;
/*printf ( "rho = %f\n", rho );*/
  for ( i = 0, ri = 1.0;  i <= BezPaletteDeg;  i++, ri *= rho )
    for ( j = 0; j < 4; j++ )
      BezPaletteCP[i][j] = BPCP[i][j]*ri;
  LoadMyPalette ();
} /*SetPaletteRho*/

char PickBezPoint1 ( double x, double y )
{
#define TOL 8.0
  int     i, k;
  float   d, dmin;
  GLfloat vp[4], q[4];

  y = (double)win2h-1.0-y;
  dmin = TOL*TOL+1.0;
  k = -1;
  for ( i = 0; i <= BezPaletteDeg; i++ ) {
    M4x4MultMVf ( vp, win2vpm, BPCP[i] );
    vp[0] = (0.5*vp[0]+0.5)*(float)win2w;
    vp[1] = (0.5*vp[1]+0.5)*(float)win2h;
    d = (vp[0]-x)*(vp[0]-x) + (vp[1]-y)*(vp[1]-y);
    if ( d < dmin ) { k = i;  dmin = d; }
  }
  if ( dmin < TOL*TOL ) {
    current = k;
    return true;
  }

  dmin = TOL*TOL+1.0;
  k = -1;
  for ( i = 0; i <= BezPaletteDeg; i++ ) {
    q[0] = (float)i/(float)BezPaletteDeg;
    q[1] = BPCP[i][2];  q[2] = 0.0;  q[3] = 1.0;
    M4x4MultMVf ( vp, win2sqvpm, q );
    vp[0] = (0.5*vp[0]+0.5)*(float)win2w;
    vp[1] = (0.5*vp[1]+0.5)*(float)win2h;
    d = (vp[0]-x)*(vp[0]-x) + (vp[1]-y)*(vp[1]-y);
    if ( d < dmin ) { k = i;  dmin = d; }
  }
  if ( dmin < TOL*TOL ) {
    current = k + MAXDEG+1;
    return true;
  }
  return false;
} /*PickBezPoint1*/

void MoveBezPoint1 ( double x, double y )
{
  GLfloat p[4], q[4];

  y = (double)win2h-1.0-y;
  p[0] = 2.0*(x/(float)win2w)-1.0;
  p[1] = 2.0*(y/(float)win2h)-1.0;
  p[2] = 0.0;  p[3] = 1.0;
  if ( current <= BezPaletteDeg ) {
    M4x4MultMVf ( q, win2vpminv, p );
    q[2] = BPCP[current][2];
    memcpy ( &BPCP[current], q, 4*sizeof(GLfloat) );
  }
  else {
    M4x4MultMVf ( q, win2sqvpminv, p );
    if ( q[1] < 0.0 ) q[1] = 0.0; else if ( q[1] > 1.0 ) q[1] = 1.0;
    BPCP[current-MAXDEG-1][2] = q[1];
  }
  SetPaletteRho ( rho );
  LoadMyPalette ();
} /*MoveBezPoint1*/

char DegreeElevation ( GLint *deg, GLfloat cp[][4], GLfloat bpcp[][4] )
{
  int   i, j, n;
  float t;

  if ( *deg < MAXDEG ) {
    n = *deg+1;
    memcpy ( cp[n], cp[n-1], 4*sizeof(GLfloat) );
    for ( i = n-1; i > 0; i-- ) {
      t = (float)i/(float)n;
      for ( j = 0; j < 4; j++ )
        cp[i][j] = t*cp[i-1][j] + (1.0-t)*cp[i][j];
    }
    *deg = n;
    for ( i = 0; i <= n; i++ ) {
      for ( j = 0; j < 3; j++ )
        bpcp[i][j] = cp[i][j]/cp[i][3];
      bpcp[i][3] = 1.0;
    }
    return true;
  }
  else
    return false;
} /*DegreeElevation*/

char DegreeReduction ( GLint *deg, GLfloat cp[][4], GLfloat bpcp[][4] )
{
  int i, j, n;

  if ( *deg > 1 ) {
    n = --(*deg);
    for ( i = 0; i <= n; i++ ) {
      for ( j = 0; j < 3; j++ )
        bpcp[i][j] = cp[i][j]/cp[i][3];
      bpcp[i][3] = 1.0;
    }
    return true;
  }
  else
    return false;
} /*DegreeReduction*/

char PaletteDegElev ( void )
{
  if ( DegreeElevation ( &BezPaletteDeg, BezPaletteCP, BPCP ) ) {
    LoadMyPalette ();
    return true;
  }
  else
    return false;
} /*PaletteDegElev*/

char PaletteDegRed ( void )
{
  if ( DegreeReduction ( &BezPaletteDeg, BezPaletteCP, BPCP ) ) {
    LoadMyPalette ();
    return true;
  }
  else
    return false;
} /*PaletteDegRed*/

char PaletteRhoChange ( char inc )
{
  if ( inc && rho < RHOMAX ) {
    SetPaletteRho ( rho*DRHO );
    return true;
  }
  else if ( !inc && rho > RHOMIN ) {
    SetPaletteRho ( rho/DRHO );
    return true;
  }
  else
    return false;
} /*PaletteRhoChange*/

