
#include <stdlib.h>
#include <stdio.h>
#include "openglheader.h"

#include "utilities.h"
#include "offsfbo.h"

static char AllocOffsTexture ( OffsFBO *fbo, int w, int h )
{
  glGenTextures ( 1, &fbo->txt );
  glBindTexture ( GL_TEXTURE_2D, fbo->txt );
  glTexStorage2D ( GL_TEXTURE_2D, 1, GL_R32F, w, h );
  glBindFramebuffer ( GL_FRAMEBUFFER, fbo->fbo );
  glFramebufferTexture ( GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, fbo->txt, 0 );
  if ( glCheckFramebufferStatus ( GL_FRAMEBUFFER ) != GL_FRAMEBUFFER_COMPLETE )
    ExitOnError ( "AllocOffsTexture" );
  glBindFramebuffer ( GL_FRAMEBUFFER, 0 );
  fbo->cmap = NewStorageBuffer ( w*h*2*sizeof(GLdouble), 0 );
  fbo->width = w;  fbo->height = h;
  ExitIfGLError ( "AllocOffsTexture" );
  return true;
} /*AllocOffsTexture*/

OffsFBO *NewOffsFBO ( int w, int h )
{
  OffsFBO *fbo;

  if ( (fbo = malloc (sizeof(OffsFBO))) ) {
    glGenFramebuffers ( 1, &fbo->fbo );
    if ( AllocOffsTexture ( fbo, w, h ) )
      return fbo;
    else {
      glDeleteFramebuffers ( 1, &fbo->fbo );
      free ( fbo );
    }
  }
  return NULL;
} /*NewOffsFBO*/

char SetOffsFBOSize ( OffsFBO *fbo, int w, int h )
{
  if ( w != fbo->width || h != fbo->height ) {
    glDeleteTextures ( 1, &fbo->txt );
    glDeleteBuffers ( 1, &fbo->cmap );
    if ( !AllocOffsTexture ( fbo, w, h ) ) {
      glDeleteFramebuffers ( 1, &fbo->fbo );
      free ( fbo );
      return false;
    }
  }
  return true;
} /*SetOffsFBOSize*/

void DeleteOffsFBO ( OffsFBO *fbo )
{
  glDeleteFramebuffers ( 1, &fbo->fbo );
  glDeleteTextures ( 1, &fbo->txt );
  glDeleteBuffers ( 1, &fbo->cmap );
  ExitIfGLError ( "DeleteOffsFBO" );
  free ( fbo );
} /*DeleteOffsFBO*/

