#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <openglheader.h>

#include "utilities.h"
#include "lights.h"

GLuint  lsbi, lsbbp;
GLint   lsbsize, lsbofs[11];

void SetLightAmbient ( LightBl *light, int l, GLfloat amb[3] )
{
  GLint ofs;

  if ( l < 0 || l >= MAX_NLIGHTS )
    return;
  memcpy ( light->ls[l].ambient, amb, 3*sizeof(GLfloat) );
  ofs = l*(lsbofs[10]-lsbofs[6]) + lsbofs[6];
  glBindBuffer ( GL_UNIFORM_BUFFER, light->lsbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, ofs, 3*sizeof(GLfloat), amb );
  ExitIfGLError ( "SetLightAmbient" );
} /*SetLightAmbient*/

void SetLightDirect ( LightBl *light, int l, GLfloat dir[3] )
{
  GLint ofs;

  if ( l < 0 || l >= MAX_NLIGHTS )
    return;
  memcpy ( light->ls[l].direct, dir, 3*sizeof(GLfloat) );
  ofs = l*(lsbofs[10]-lsbofs[6]) + lsbofs[7];
  glBindBuffer ( GL_UNIFORM_BUFFER, light->lsbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, ofs, 3*sizeof(GLfloat), dir );
  ExitIfGLError ( "SetLightDirect" );
} /*SetLightDirect*/

void SetLightPosition ( LightBl *light, int l, GLfloat pos[4] )
{
  GLint ofs;

  if ( l < 0 || l >= MAX_NLIGHTS )
    return;
  memcpy ( light->ls[l].position, pos, 4*sizeof(GLfloat) );
  ofs = l*(lsbofs[10]-lsbofs[6]) + lsbofs[8];
  glBindBuffer ( GL_UNIFORM_BUFFER, light->lsbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, ofs, 4*sizeof(GLfloat), pos );
  ExitIfGLError ( "SetLightPosition" );
} /*SetLightPosition*/

void SetLightAttenuation ( LightBl *light, int l, GLfloat atn[3] )
{
  GLint ofs;

  if ( l < 0 || l >= MAX_NLIGHTS )
    return;
  memcpy ( light->ls[l].attenuation, atn, 3*sizeof(GLfloat) );
  ofs = l*(lsbofs[10]-lsbofs[6]) + lsbofs[9];
  glBindBuffer ( GL_UNIFORM_BUFFER, light->lsbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, ofs, 3*sizeof(GLfloat), atn );
  ExitIfGLError ( "SetLightAttenuation" );
} /*SetLightAttenuation*/

void SetLightOnOff ( LightBl *light, int l, char on )
{
  GLuint mask;

  if ( l < 0 || l >= MAX_NLIGHTS )
    return;
  mask = 0x01 << l;
  if ( on ) {
    light->mask |= mask;
    if ( l >= light->nls )
      light->nls = l+1;
  }
  else {
    light->mask &= ~mask;
    for ( mask = 0x01 << (light->nls-1); mask; mask >>= 1 ) {
      if ( light->mask & mask )
        break;
      else
        light->nls --;
    }
  }
  glBindBuffer ( GL_UNIFORM_BUFFER, light->lsbuf );
  glBufferSubData ( GL_UNIFORM_BUFFER, lsbofs[0], sizeof(GLuint), &light->nls );
  glBufferSubData ( GL_UNIFORM_BUFFER, lsbofs[1], sizeof(GLuint), &light->mask );
  ExitIfGLError ( "SetLightOnOff" );
} /*SetLightOnOff*/

void SetHemisphericalLight ( LightBl *light, GLfloat zenith[3],
                             GLfloat sky[4], GLfloat ground[4] )
{
  glBindBuffer ( GL_UNIFORM_BUFFER, light->lsbuf );
  V3Normalisef ( zenith );
  memcpy ( light->hl.zenith, zenith, 3*sizeof(GLfloat) );
  memcpy ( light->hl.sky, sky, 4*sizeof(GLfloat) );
  memcpy ( light->hl.ground, ground, 4*sizeof(GLfloat) );
  glBufferSubData ( GL_UNIFORM_BUFFER, lsbofs[3], 3*sizeof(GLfloat), zenith );
  glBufferSubData ( GL_UNIFORM_BUFFER, lsbofs[4], 4*sizeof(GLfloat), sky );
  glBufferSubData ( GL_UNIFORM_BUFFER, lsbofs[5], 4*sizeof(GLfloat), ground );
  ExitIfGLError ( "SetHemisphericalLight" );
} /*SetHemisphericalLight*/

void SetHemisphericalLightOnOff ( LightBl *light, char on )
{
  glBindBuffer ( GL_UNIFORM_BUFFER, light->lsbuf );
  light->hlen = on != 0;
  glBufferSubData ( GL_UNIFORM_BUFFER, lsbofs[2], sizeof(GLuint), &light->hlen );
  ExitIfGLError ( "SetHemisphericalLightOnOff" );
} /*SetHemisphericalLightOnOff*/

