#version 420

#define PROJ_LINEAR          0
#define PROJ_POINT_PANORAMA  1
#define PROJ_LINEAR_PANORAMA 2
#define PROJ_SPHERICAL       3

#define TR_MAT_INDEX(i,j) ( (i)*((i)+1)/2+(j) )
  
layout(triangles) in;
layout(triangle_strip,max_vertices=48) out;

in Vertex {
  vec4 Colour;
} In[];

out NVertex {
  vec4 eyepos;
  vec4 Colour;
  vec3 Position;
  vec3 Normal;
} Out;

uniform TransBlock {
  mat4 mm, mmti, vm, pm;
  vec4 eyepos;
} trb;

uniform NLProjection {
  int   type;
  float tantheta;
} nlp;

vec4 NonlinTransformation ( vec4 p )
{
  float R;

  switch ( nlp.type ) {
case PROJ_POINT_PANORAMA:
    R = length ( p.xz );
    return vec4 ( atan ( p.x, -p.z ), p.y/R, -R, p.w );
case PROJ_LINEAR_PANORAMA:
    R = length ( p.xz );
    return vec4 ( atan ( p.x, -p.z ), p.y + R*nlp.tantheta, -R, p.w );
case PROJ_SPHERICAL:
    R = length ( p.xyz );
    return vec4 ( p.x/R, p.y/R, -R/p.w, 1.0 );
default:
    return p;
  }
} /*NonlinTransformation*/

vec3 nv;
vec4 ppp[28], ccc[28], epos[28];
vec3 pos[28];

void Emit ( int i, int j )
{
  int k;

  k = TR_MAT_INDEX ( i, j );
  gl_Position = ppp[k];
  Out.eyepos = epos[k];
  Out.Position = pos[k];
  Out.Normal = nv;
  Out.Colour = ccc[k];
  EmitVertex ();
} /*Emit*/

void main ( void )
{
  int  i, j, k;
  vec3 v1, v2;
  vec4 p, q;
  float x, y, z;

  v1 = (trb.mm*(gl_in[1].gl_Position - gl_in[0].gl_Position)).xyz;
  v2 = (trb.mm*(gl_in[2].gl_Position - gl_in[0].gl_Position)).xyz;
  nv = normalize ( cross ( v1, v2 ) );
  if ( nlp.type == PROJ_LINEAR ) {
    for ( i = 0; i < 3; i++ ) {
      p = trb.mm*gl_in[i].gl_Position;
      gl_Position = trb.pm*(NonlinTransformation ( trb.vm*p ));
      Out.eyepos = trb.eyepos;
      Out.Position = p.xyz/p.w;
      Out.Normal = nv;
      Out.Colour = In[i].Colour;
      EmitVertex ();
    }
    EndPrimitive ();
  }
  else {
    for ( i = k = 0;  i < 7; i++ ) {
      x = float(6-i)/6.0;
      for ( j = 0;  j <= i;  j++, k++ ) {
        y = float(i-j)/6.0;
        z = 1.0 - x - y;
        p = x*gl_in[0].gl_Position + y*gl_in[1].gl_Position +
            z*gl_in[2].gl_Position;
        p = trb.mm * p;
        pos[k] = p.xyz/p.w;
        q = NonlinTransformation ( trb.vm*p );
        ppp[k] = trb.pm * q;
        if ( nlp.type == PROJ_LINEAR_PANORAMA ) {
          q.x = q.z = 0.0;
          epos[k] = inverse ( trb.vm )*q;
        }
        else
          epos[k] = trb.eyepos;
        ccc[k] = x*In[0].Colour + y*In[1].Colour + z*In[2].Colour;
      }
    }
    for ( i = 1; i < 7; i++ ) {
      Emit ( i, 0 );
      for ( j = 0; j < i; j++ ) {
        Emit ( i-1, j );
        Emit ( i, j+1 );
      }
      EndPrimitive ();
    }
  }
} /*main*/

