#version 440

#define A 0.57735027  /* sqrt(1/3) */

layout(points) in;
layout(triangle_strip,max_vertices=12) out;

in int instanceID[];

out FVertex {
    vec3 Colour, Position, Normal, TNormal;
  } Out;

uniform TransBlock {
    mat4 mm, mmti, vm, pm, vpm;
    vec4 eyepos;
  } trb;

uniform int level;

const int t[21][3] =
  {{0,0,0},{1,0,0},{2,0,0},{0,1,0},{2,1,0},{0,2,0,},{1,2,0},{2,2,0},
   {0,0,1},{2,0,1},{0,2,1},{2,2,1},
   {0,0,2},{1,0,2},{2,0,2},{0,1,2},{2,1,2},{0,2,2,},{1,2,2},{2,2,2},
   {0,0,0}};
const int ip0[4] = {0,3,4,7}, ip1[4] = {1,5,2,6}, ip2[4] = {1,0,5,4},
          ip3[4] = {2,6,3,7}, ip4[4] = {1,2,0,3}, ip5[4] = {5,4,6,7};

vec3 p[8];

void OutputVertex ( vec3 p, vec3 nv )
{
  vec4 q;

  Out.Colour = vec3( p.x/(2*A)+0.5, p.y/(2.0*A)+0.5, p.z/(2.0*A)+0.5 );
  Out.Position = (q = trb.mm*vec4(p,1.0)).xyz;
  Out.Normal = Out.TNormal = mat3(trb.mmti) * nv;
  gl_Position = trb.vpm * q;
  EmitVertex ();
} /*OutputVertex*/

void OutputFacet ( int ip[4], vec3 nv )
{
  int i;

  for ( i = 0; i < 4; i++ )
    OutputVertex ( p[ip[i]], nv );
  EndPrimitive ();
} /*OutputFacet*/

void main ( void )
{
  float p0[3], p1[3], a;
  vec4  epm;
  int   i, j, k, l;

  p0[0] = p0[1] = p0[2] = -A;  p1[0] = p1[1] = p1[2] = A;
  for ( i = 0, j = instanceID[0];  i < level;  i++, j /= 20 ) {
    k = j % 20;
    for ( l = 0; l < 3; l++ )
      switch ( t[k][l] ) {
    case 0:
        p1[l] = (2.0*p0[l] + p1[l])/3.0;
        break;
    case 1:
        a = (2.0*p0[l] + p1[l])/3.0;
        p1[l] = (p0[l] + 2.0*p1[l])/3.0;
        p0[l] = a;
        break;
    case 2:
        p0[l] = (p0[l] + 2.0*p1[l])/3.0;
        break;
      }
  }
  p[0] = vec3(p0[0],p0[1],p0[2]);  p[1] = vec3(p1[0],p0[1],p0[2]);
  p[2] = vec3(p1[0],p1[1],p0[2]);  p[3] = vec3(p0[0],p1[1],p0[2]);
  p[4] = vec3(p0[0],p0[1],p1[2]);  p[5] = vec3(p1[0],p0[1],p1[2]);
  p[6] = vec3(p1[0],p1[1],p1[2]);  p[7] = vec3(p0[0],p1[1],p1[2]);
  epm = transpose(trb.mmti) * trb.eyepos;
  if ( level == 0 ) k = 20;
  if ( t[k][0] != 1 && epm.x - p0[0] < 0.0 )
    OutputFacet ( ip0, vec3(-1.0,0.0,0.0) );
  if ( t[k][0] != 1 && epm.x - p1[0] > 0.0 )
    OutputFacet ( ip1, vec3(1.0,0.0,0.0) );
  if ( t[k][1] != 1 && epm.y - p0[1] < 0.0 )
    OutputFacet ( ip2, vec3(0.0,-1.0,0.0) );
  if ( t[k][1] != 1 && epm.y - p1[1] > 0.0 )
    OutputFacet ( ip3, vec3(0.0,1.0,0.0) );
  if ( t[k][2] != 1 && epm.z - p0[2] < 0.0 )
    OutputFacet ( ip4, vec3(0.0,0.0,-1.0) );
  if ( t[k][2] != 1 && epm.z - p1[2] > 0.0 )
    OutputFacet ( ip5, vec3(0.0,0.0,1.0) );
} /*main*/

